"""Schema validation pre-processor."""

import typing

from ... import exceptions as _exceptions
from ... import types as _oa_types
from .. import helpers as _helpers
from . import association
from . import model
from . import property_
from . import schemas_validation
from . import spec_validation
from . import types
from . import unique_secondary
from . import unique_tablename
from . import unmanaged


def _get_properties_results(
    schemas: _oa_types.Schemas, schema: _oa_types.Schema
) -> typing.Iterable[typing.Tuple[str, types.Result]]:
    """Get an iterator with properties results."""
    # Get model properties
    properties = _helpers.iterate.properties_items(
        schema=schema, schemas=schemas, stay_within_model=True
    )
    # Filter any keys that are not string
    properties = filter(lambda args: isinstance(args[0], str), properties)
    # Check properties
    return map(
        lambda args: (args[0], property_.check(schemas, schema, args[0], args[1])),
        properties,
    )


def _process_model(
    schemas: _oa_types.Schemas, schema_name: str, schema: _oa_types.Schema
) -> None:
    """
    Validate the model schema properties.

    Assume the schema is valid at the model level.

    Args:
        schemas: All defined schemas used to resolve any $ref.
        schema_name: The name of the schema to validate.
        schema: The schema to validate.

    """
    properties_results = _get_properties_results(schemas, schema)
    invalid_properties_result = next(
        filter(lambda args: not args[1].valid, properties_results), None
    )
    if invalid_properties_result is not None:
        name, result = invalid_properties_result
        raise _exceptions.MalformedSchemaError(
            f"{schema_name} :: {name} :: {result.reason}"
        )


def _other_schemas_checks(*, schemas: _oa_types.Schemas) -> types.Result:
    """
    Check that at least 1 model is defined and for multiple tablename.

    Args:
        schemas: All defined schemas.

    Returns:
        Whether the schemas are valid with a reason if they are not.

    """
    one_model_result = check_one_model(schemas=schemas)
    if not one_model_result.valid:
        return one_model_result

    unique_tablename_result = unique_tablename.check(schemas=schemas)
    if not unique_tablename_result.valid:
        return unique_tablename_result

    unique_secondary_result = unique_secondary.check(schemas=schemas)
    if not unique_secondary_result.valid:
        return unique_secondary_result

    association_result = association.check(schemas=schemas)
    if not association_result.valid:
        return association_result

    return types.Result(valid=True, reason=None)


def process(*, schemas: _oa_types.Schemas) -> None:
    """
    Validate schemas.

    Args:
        schemas: The schemas to validate.

    """
    schemas_result = schemas_validation.check(schemas=schemas)
    if not schemas_result.valid:
        raise _exceptions.MalformedSchemaError(schemas_result.reason)

    # Check constructable schemas model
    constructables = _helpers.iterate.constructable(schemas=schemas)
    model_results = map(
        lambda args: (args[0], model.check(schemas, args[1])), constructables
    )
    invalid_model_result = next(
        filter(lambda args: not args[1].valid, model_results), None
    )
    if invalid_model_result is not None:
        name, result = invalid_model_result
        raise _exceptions.MalformedSchemaError(f"{name} :: {result.reason}")

    # Check constructable schemas properties
    constructables = _helpers.iterate.constructable(schemas=schemas)
    for constructable in constructables:
        name, schema = constructable
        _process_model(schemas, name, schema)

    other_results_result = _other_schemas_checks(schemas=schemas)
    if not other_results_result.valid:
        raise _exceptions.MalformedSchemaError(other_results_result.reason)


def check_one_model(*, schemas: _oa_types.Schemas) -> types.Result:
    """
    Check that there is at least 1 model in the schemas.

    Args:
        schemas: The schemas to validate.

    Returns:
        Whether the schemas contain at least 1 model.

    """
    constructables = _helpers.iterate.constructable(schemas=schemas)
    if not any(constructables):
        return types.Result(
            False,
            "specification must define at least 1 schema with the x-tablename key",
        )

    return types.Result(True, None)


def _check_model_properties(
    *, schemas: _oa_types.Schemas, schema: _oa_types.Schema
) -> types.TProperties:
    """
    Check the properties of a model.

    Args:
        schemas: All defined schemas used to resolve any $ref.
        schema_name: The name of the schema to validate.
        schema: The schema to validate.

    Returns:
        Whether the properties are valid.

    """
    properties_results = _get_properties_results(schemas, schema)
    properties_t_results: typing.Iterable[typing.Tuple[str, types.TProperty]] = map(
        lambda args: (args[0], {"result": types.t_result_from_result(args[1])}),
        properties_results,
    )
    return dict(properties_t_results)


def _check_model(schemas: _oa_types.Schemas, schema: _oa_types.Schema) -> types.TModel:
    """
    Check a model.

    Args:
        schema: The schema of the model to check.

    Returns:
        Whether the model and its properties are valid with a reason if it is not.

    """
    model_result = model.check(schemas, schema)
    return {
        "result": types.t_result_from_result(model_result),
        "properties": _check_model_properties(schemas=schemas, schema=schema),
    }


def check_models(*, schemas: _oa_types.Schemas) -> types.TModels:
    """
    Check the models of a schema.

    Assume the schemas is valid although any of its models may not.

    Args:
        schemas: The schemas to check.

    Returns:
        The result for each model.

    """
    constructables = _helpers.iterate.constructable(schemas=schemas)
    constructables_result = map(
        lambda args: (args[0], _check_model(schemas, args[1])), constructables
    )
    return dict(constructables_result)


def check(*, spec: typing.Any) -> types.TSpec:
    """
    Check a specification.

    Args:
        spec: The specification to check.

    Returns:
        Whether the specification is valid with a reason if it is not.

    """
    # Check spec to schemas
    spec_result = spec_validation.check(spec=spec)
    if not spec_result.valid:
        return {"result": types.t_result_from_result(spec_result)}

    # Check that there is at least 1 model
    assert isinstance(spec, dict)
    components = spec.get("components")
    assert isinstance(components, dict)
    schemas = components.get("schemas")
    assert isinstance(schemas, dict)

    try:
        other_result = _other_schemas_checks(schemas=schemas)
        if not other_result.valid:
            return {"result": types.t_result_from_result(other_result)}
    except _exceptions.BaseError:
        # Something else is wrong
        pass

    return {"result": {"valid": True}, "models": check_models(schemas=schemas)}



"""Retrieve mixin classes."""

import importlib
import inspect
import typing

from . import exceptions


def _import_mixin(mixin: str) -> typing.Type:
    """Import the mixin."""
    assert "." in mixin
    module_name, class_name = mixin.rsplit(".", 1)

    try:
        module = importlib.import_module(module_name)
    except ImportError as exc:
        raise exceptions.MalformedExtensionPropertyError(
            f'error when attempting to import "{module_name}"'
        ) from exc

    try:
        class_ = getattr(module, class_name)
    except AttributeError as exc:
        raise exceptions.MalformedExtensionPropertyError(
            f'the module "{module_name}" does not have the "{class_name}" class'
        ) from exc

    if not inspect.isclass(class_):
        raise exceptions.MalformedExtensionPropertyError(
            f'"{class_name}" is not a class'
        )

    return class_


def get(*, mixins: typing.List[str]) -> typing.Tuple[typing.Type, ...]:
    """
    Import the mixins.

    Assume that the mixins are valid instances of x-mixins.

    Raise MalformedExtensionPropertyError if the value of x-mixins cannot be imported
    or is not a class.

    Algorithm:
    1. split each mixin on the last dot,
    2. use the first part to import the module,
    3. use the second part to retrieve the class and
    4. check that the class is actually a class.

    Args:
        mixins: The mixins to load.

    Returns:
        The mixin classes.

    """
    return tuple(map(_import_mixin, mixins))



"""Dictionary conversion for object."""

import typing

from ... import exceptions
from ... import helpers
from ... import types as oa_types
from .. import types


def _convert_relationship(*, value: types.TModel) -> types.TOptObjectDict:
    """
    Convert object relationship property to a dictionary.

    Raises InvalidModelInstanceError if the value does not have a to_dict function.
    Raises InvalidModelInstanceError if the value has a to_dict function that expects
        arguments.

    Args:
        value: The value to convert.

    Returns:
        The object as a dictionary.

    """
    if value is None:
        return None

    try:
        return value.to_dict()
    except AttributeError as exc:
        raise exceptions.InvalidModelInstanceError(
            "The object property instance does not have a to_dict implementation."
        ) from exc
    except TypeError as exc:
        raise exceptions.InvalidModelInstanceError(
            "The object property instance to_dict implementation is "
            "expecting arguments."
        ) from exc


def _convert_read_only(
    *, schema: oa_types.Schema, value: typing.Any
) -> types.TOptObjectDict:
    """
    Convert readOnly value to a dictionary.

    Raise MalformedSchemaError if the schema does not have properties.
    Raise MalformedSchemaError if the schema has empty properties.
    """
    if value is None:
        return None

    properties = schema.get("properties")
    if properties is None:
        raise exceptions.MalformedSchemaError(
            "readOnly object definition must have properties."
        )
    if not isinstance(properties, dict):
        raise exceptions.MalformedSchemaError(
            "readOnly object definition must have dictionary properties."
        )
    if not properties:
        raise exceptions.MalformedSchemaError(
            "readOnly object definitions must have at least 1 property."
        )
    return_dict = {}
    for key in properties.keys():
        return_dict[key] = getattr(value, key, None)
    return return_dict


def convert(
    value: typing.Any,
    *,
    schema: oa_types.Schema,
    read_only: typing.Optional[bool] = None,
) -> types.TOptObjectDict:
    """
    Convert object schema value to dictionary.

    Args:
        value: The value to convert.
        schema: The schema for the value.
        read_only (optional): Whether the schema is read only.

    """
    schema_read_only = helpers.peek.read_only(schema=schema, schemas={})
    if read_only or schema_read_only:
        return _convert_read_only(schema=schema, value=value)
    return _convert_relationship(value=value)



"""Helper functions."""
# pylint: disable=useless-import-alias

from . import all_of as all_of
from . import command
from . import ext_prop as ext_prop
from . import foreign_key as foreign_key
from . import inheritance as inheritance
from . import oa_to_py_type as oa_to_py_type
from . import peek as peek
from . import property_ as property_
from . import ref as ref
from . import relationship as relationship
from . import schema as schema
from . import type_ as type_
from .calculate_nullable import calculate_nullable as calculate_nullable
from .define_all import define_all as define_all



"""Convert array values to columns."""

import functools

from ... import exceptions
from ... import helpers
from ... import types as oa_types
from .. import types
from . import object_


def convert(
    value: types.TOptArrayDict, *, schema: oa_types.Schema
) -> types.TOptArrayCol:
    """
    Convert array value from a dictionary to a column.

    Raises MalformedSchemaError if the items schema is missing from the schema.
    Raises MalformedSchemaError if the items type is not object.
    Raises InvalidInstanceError if the value is not an iterable.

    Args:
        value: The value to convert.
        schema: The schema of the value.

    Returns:
        The converted value.

    """
    # Check the schema
    items_schema = schema.get("items")
    if items_schema is None:
        raise exceptions.MalformedSchemaError(
            "To construct array parameters the schema for the property "
            "must include the items property with the information about "
            "the array items."
        )
    items_type = helpers.peek.type_(schema=items_schema, schemas={})
    if items_type != "object":
        raise exceptions.MalformedSchemaError(
            "The type of the array items must be object."
        )

    if value is None:
        return None
    # Convert values
    item_conversion = functools.partial(object_.convert, schema=items_schema)
    try:
        converted_items = map(item_conversion, value)
    except TypeError as exc:
        raise exceptions.InvalidInstanceError("Array values must be iterable.") from exc
    return list(converted_items)



"""Calculate the column artifacts for a model."""

import itertools
import typing

from open_alchemy import schemas
from open_alchemy import types as oa_types

from .. import types
from . import type_


def calculate(
    *, artifacts: schemas.artifacts.types.ModelArtifacts
) -> typing.List[types.ColumnArtifacts]:
    """
    Calculate the column artifacts from model schema artifacts.

    Args:
        artifacts: The schema artifacts for a model.

    Returns:
        The artifacts for the columns.

    """
    # Process properties
    no_backref_properties = filter(
        lambda args: args[1].type != oa_types.PropertyType.BACKREF,
        artifacts.properties,
    )
    no_backref_no_dict_ignore_properties = filter(
        lambda args: not (
            args[1].type == oa_types.PropertyType.SIMPLE
            and args[1].extension.dict_ignore
        ),
        no_backref_properties,
    )
    properties_columns = map(
        lambda args: types.ColumnArtifacts(
            name=args[0],
            type=type_.model(artifacts=args[1]),
            description=args[1].description,
        ),
        no_backref_no_dict_ignore_properties,
    )

    # Process back references
    backrefs_columns = map(
        lambda args: types.ColumnArtifacts(
            name=args[0],
            type=f'typing.Optional["T{args[1].child}"]'
            if args[1].type == schemas.artifacts.types.BackrefSubType.OBJECT
            else f'typing.Sequence["T{args[1].child}"]',
            description=None,
        ),
        artifacts.backrefs,
    )

    return list(itertools.chain(properties_columns, backrefs_columns))



"""Helpers to process schemas."""

import itertools
import typing

from ... import types
from . import iterate


class TArtifacts(typing.NamedTuple):
    """The return value of _calculate_schema."""

    schema_name: str
    property_name: str
    property_schema: typing.Any


TArtifactsIter = typing.Iterator[TArtifacts]
TArtifactsGroupedIter = typing.Iterator[typing.Tuple[str, TArtifactsIter]]
TSchemaIter = typing.Iterable[typing.Tuple[str, types.Schema]]
TGetSchemaArtifacts = typing.Callable[
    [types.Schemas, str, types.Schema], TArtifactsIter
]


def get_artifacts(
    *, schemas: types.Schemas, get_schema_artifacts: TGetSchemaArtifacts
) -> TArtifactsIter:
    """
    Get all artifacts information from the schemas.

    Assume schemas is valid.

    Takes all schemas, retrieves all constructable schemas, for each schema retrieves
    all artifacts and returns an iterable with all the captured back references.

    Args:
        schemas: The schemas to process.
        get_schema_artifacts: Function that retrieves the artifacts for the schema.

    Returns:
        All backreference information.

    """
    # Retrieve all constructable schemas
    constructables = iterate.constructable(schemas=schemas)
    # Retrieve all artifacts
    artifacts_iters = map(
        lambda args: get_schema_artifacts(schemas, *args), constructables
    )
    # Unpack nested iterators
    return itertools.chain(*artifacts_iters)


TOutput = typing.TypeVar("TOutput")
TOutputIter = typing.Iterator[typing.Tuple[str, TOutput]]
TCalculateOutput = typing.Callable[[TArtifactsIter], TOutput]


def calculate_outputs(
    *, artifacts: TArtifactsIter, calculate_output: TCalculateOutput
) -> TOutputIter:
    """
    Convert artifacts iterator to an output iterator.

    Algorithm:
    1. sort and group by schema name and
    2. call calculate_output on the grouped artifacts.

    Args:
        artifacts: The artifacts to convert.
        calculate_output: Calculate the output from artifacts for a schema.

    Returns:
        An iterator with the converted output.

    """
    # Sort and group
    sorted_artifacts = sorted(artifacts, key=lambda backref: backref.schema_name)
    grouped_artifacts = itertools.groupby(
        sorted_artifacts, lambda backref: backref.schema_name
    )
    # Map to output
    return map(lambda args: (args[0], calculate_output(args[1])), grouped_artifacts)



"""Define all the models with x-tablename properties."""

from .. import types
from . import inheritance as inheritance_helper
from . import schema as schema_helper


def define_all(*, model_factory: types.ModelFactory, schemas: types.Schemas) -> None:
    """
    Define all the models with x-tablename properties.

    Args:
        model_factory: Factory used to construct models.
        schemas: The schemas from which to define all.

    """
    for name, schema in schemas.items():
        if not schema_helper.constructable(schema=schema, schemas=schemas):
            continue
        if schema_helper.inherits(schema=schema, schemas=schemas):
            parents = inheritance_helper.get_parents(schema=schema, schemas=schemas)
            for parent in parents:
                model_factory(name=parent)
        model_factory(name=name)



"""Convert array to dictionary."""

import functools
import typing

from ... import exceptions
from ... import helpers
from ... import types as ao_types
from .. import types
from . import object_


def convert(value: typing.Any, *, schema: ao_types.Schema) -> types.TOptArrayDict:
    """
    Convert array property so that it can be included in an object dictionary.

    Raises MalformedSchemaError if schema does not define item schema.
    Raises MalformedSchemaError if the item schema is not of type object.
    Raises MalformedSchemaError if the item schema is not of type object.

    Args:
        value: The value to convert.
        schema: The schema for the value.

    Returns:
        The value converted to a list of dictionary.

    """
    if value is None:
        return None
    item_schema = schema.get("items")
    if item_schema is None:
        raise exceptions.MalformedSchemaError(
            "The array item schema must have an items property."
        )
    item_type = helpers.peek.type_(schema=item_schema, schemas={})
    if item_type != "object":
        raise exceptions.FeatureNotImplementedError(
            "The array item schema must be of type object."
        )
    read_only = helpers.peek.read_only(schema=schema, schemas={})
    item_conversion = functools.partial(
        object_.convert, schema=item_schema, read_only=read_only
    )
    try:
        converted_items = map(item_conversion, value)
    except TypeError as exc:
        raise exceptions.InvalidInstanceError("Array values must be iterable.") from exc
    return list(converted_items)



"""Pre-process schemas by adding any back references into the schemas."""

import functools
import typing

from .. import helpers as oa_helpers
from .. import types
from . import helpers


class TArtifacts(helpers.process.TArtifacts):
    """The return value of _calculate_schema."""

    property_schema: types.Schema


def _calculate_artifacts(
    schema_name: str, schemas: types.Schemas, schema: types.Schema
) -> TArtifacts:
    """
    Calculate the artifacts for the schema for a back reference.

    Args:
        schema_name: The name of the schema that the property is on.
        schema: All the defines schemas.
        schema: The schema of a property with a back reference.

    Returns:
        The name of the schema that is being referenced, the name of the property to be
        added and the schema for the back reference.

    """
    is_array: bool = False
    ref: typing.Optional[str]
    backref: typing.Optional[str]

    # Handle array
    items_schema = oa_helpers.peek.items(schema=schema, schemas=schemas)
    if items_schema is not None:
        if oa_helpers.peek.secondary(schema=items_schema, schemas=schemas) is not None:
            is_array = True

        ref = oa_helpers.peek.ref(schema=items_schema, schemas=schemas)
        backref = oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.backref, schema=items_schema, schemas=schemas
        )
    # Handle object
    else:
        uselist: typing.Optional[bool] = oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.uselist, schema=schema, schemas=schemas
        )
        if uselist is not False:
            is_array = True
        ref = oa_helpers.peek.ref(schema=schema, schemas=schemas)
        backref = oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.backref, schema=schema, schemas=schemas
        )

    # Resolve name
    assert ref is not None
    ref_schema_name, _ = oa_helpers.ref.get_ref(ref=ref, schemas=schemas)

    # Calculate schema
    assert backref is not None
    return_schema: types.Schema = {"type": "object", "x-de-$ref": schema_name}
    if is_array:
        return_schema = {"type": "array", "items": return_schema}

    return TArtifacts(ref_schema_name, backref, return_schema)


def _get_schema_backrefs(
    schemas: types.Schemas,
    schema_name: str,
    schema: types.Schema,
) -> helpers.process.TArtifactsIter:
    """
    Get the backrefs for a schema.

    Takes a constructable schema, gets all properties, filters for those that define a
    backref and retrieves the information to define a back reference.

    Args:
        schemas: All schemas.
        schema_name: The name of the schema.
        schema: A constructable schema.

    Returns:
        An iterable with all back references for the schema.

    """
    # Get all the properties of the schema
    names_properties = helpers.iterate.properties_items(
        schema=schema, schemas=schemas, stay_within_model=True
    )
    # Remove property name
    properties = map(lambda arg: arg[1], names_properties)
    # Remove properties that don't define back references
    defines_backref_schemas = functools.partial(helpers.backref.defined, schemas)
    backref_properties = filter(defines_backref_schemas, properties)
    # Capture information for back references
    calculate_artifacts_schema_name_schemas = functools.partial(
        _calculate_artifacts, schema_name, schemas
    )
    return map(calculate_artifacts_schema_name_schemas, backref_properties)


def _backrefs_to_schema(backrefs: helpers.process.TArtifactsIter) -> types.Schema:
    """
    Convert to the schema with the x-backrefs value from backrefs.

    Args:
        backrefs: The back references to convert.

    Returns:
        The schema with the x-backrefs value.

    """
    return {
        "type": "object",
        "x-backrefs": {property_name: schema for _, property_name, schema in backrefs},
    }


def process(*, schemas: types.Schemas) -> None:
    """
    Pre-process the schemas to add back references as required.

    These are added as allOf with x-backrefs which include the property name and schema
    for the property.

    Args:
        schemas: The schemas to process.

    """
    # Retrieve back references
    backrefs = helpers.process.get_artifacts(
        schemas=schemas, get_schema_artifacts=_get_schema_backrefs
    )
    # Map to a schema for each grouped back references
    backref_schemas = helpers.process.calculate_outputs(
        artifacts=backrefs, calculate_output=_backrefs_to_schema
    )
    # Convert to list to resolve iterator
    backref_schema_list = list(backref_schemas)
    # Add backreferences to schemas
    for name, backref_schema in backref_schema_list:
        schemas[name] = {"allOf": [schemas[name], backref_schema]}



"""Artifacts for the arguments."""

import itertools
import json
import typing

from open_alchemy import helpers
from open_alchemy import schemas
from open_alchemy import types as oa_types

from .. import types
from . import type_


class ReturnValue(typing.NamedTuple):
    """The return value for the calculated artifacts."""

    required: typing.List[types.ColumnArgArtifacts]
    not_required: typing.List[types.ColumnArgArtifacts]


def _get_read_only(
    artifacts: schemas.artifacts.types.TAnyPropertyArtifacts,
) -> typing.Optional[bool]:
    """Get read only for any property artifacts."""
    if artifacts.type == oa_types.PropertyType.SIMPLE:
        return artifacts.open_api.read_only
    if artifacts.type == oa_types.PropertyType.JSON:
        return artifacts.open_api.read_only
    return None


def _map_default(
    *, artifacts: schemas.artifacts.types.TAnyPropertyArtifacts
) -> oa_types.TColumnDefault:
    """
    Map default value from OpenAPI to be ready to be inserted into a Python file.

    First applies JSON formatting for a string and then using the type mapping helper
    function.

    Args:
        artifacts: The artifacts from which to map the default value.

    Returns:
        The mapped default value.

    """
    if artifacts.type != oa_types.PropertyType.SIMPLE:
        return None

    default = artifacts.open_api.default
    if default is None:
        return None

    # Escape string
    if artifacts.open_api.type == "string" and artifacts.open_api.format not in {
        "date",
        "date-time",
    }:
        default = json.dumps(default)

    # Handle bytes
    if artifacts.open_api.type == "string" and artifacts.open_api.format == "binary":
        return f"b{default}"

    # Map type
    mapped_default = helpers.oa_to_py_type.convert(
        value=default, type_=artifacts.open_api.type, format_=artifacts.open_api.format
    )

    # Get the repr if it isn't a float, bool, number nor str
    if isinstance(mapped_default, (int, float, str, bool)):
        return mapped_default
    return repr(mapped_default)


def _calculate(
    *,
    artifacts: typing.Iterable[
        typing.Tuple[str, schemas.artifacts.types.TAnyPropertyArtifacts]
    ],
) -> typing.Iterable[types.ColumnArgArtifacts]:
    """Calculate the arg artifacts from property artifacts."""
    no_backref_properties = filter(
        lambda args: args[1].type != oa_types.PropertyType.BACKREF,
        artifacts,
    )
    no_backref_no_dict_ignore_properties = filter(
        lambda args: not (
            args[1].type == oa_types.PropertyType.SIMPLE
            and args[1].extension.dict_ignore
        ),
        no_backref_properties,
    )
    return map(
        lambda args: types.ColumnArgArtifacts(
            name=args[0],
            init_type=type_.arg_init(artifacts=args[1]),
            from_dict_type=type_.arg_from_dict(artifacts=args[1]),
            default=_map_default(artifacts=args[1]),
            read_only=_get_read_only(artifacts=args[1]),
        ),
        no_backref_no_dict_ignore_properties,
    )


def calculate(*, artifacts: schemas.artifacts.types.ModelArtifacts) -> ReturnValue:
    """
    Calculate the args artifacts from model schema artifacts.

    Args:
        artifacts: The schema artifacts for a model.

    Returns:
        The artifacts for the args.

    """
    required = filter(lambda args: args[1].required, artifacts.properties)
    not_required = filter(lambda args: not args[1].required, artifacts.properties)

    # Process back references
    backrefs_columns = map(
        lambda args: types.ColumnArgArtifacts(
            name=args[0],
            init_type=f'typing.Optional["T{args[1].child}"]'
            if args[1].type == schemas.artifacts.types.BackrefSubType.OBJECT
            else f'typing.Optional[typing.Sequence["T{args[1].child}"]]',
            from_dict_type="",
            read_only=True,
        ),
        artifacts.backrefs,
    )

    return ReturnValue(
        required=list(_calculate(artifacts=required)),
        not_required=list(
            itertools.chain(_calculate(artifacts=not_required), backrefs_columns)
        ),
    )



"""Generate source code for a model."""

import os

import jinja2

from .. import types

_DIRECTORY = os.path.dirname(__file__)
# SQLAlchemy template
_SQLALCHEMY_TEMPLATE_FILENAME = os.path.join(_DIRECTORY, "sqlalchemy.j2")
with open(_SQLALCHEMY_TEMPLATE_FILENAME) as in_file:
    _SQLALCHEMY_TEMPLATE = in_file.read()
# TypedDict required template
_TYPED_DICT_REQUIRED_TEMPLATE_FILENAME = os.path.join(
    _DIRECTORY, "typed_dict_required.j2"
)
with open(_TYPED_DICT_REQUIRED_TEMPLATE_FILENAME) as in_file:
    _TYPED_DICT_REQUIRED_TEMPLATE = in_file.read()
# TypedDict not required template
_TYPED_DICT_NOT_REQUIRED_TEMPLATE_FILENAME = os.path.join(
    _DIRECTORY, "typed_dict_not_required.j2"
)
with open(_TYPED_DICT_NOT_REQUIRED_TEMPLATE_FILENAME) as in_file:
    _TYPED_DICT_NOT_REQUIRED_TEMPLATE = in_file.read()
# Overall template
_TEMPLATE_FILENAME = os.path.join(_DIRECTORY, "template.j2")
with open(_TEMPLATE_FILENAME) as in_file:
    _TEMPLATE = in_file.read()


def sqlalchemy(*, artifacts: types.SQLAlchemyModelArtifacts) -> str:
    """
    Generate the SQLAlchemy model source code.

    Args:
        artifacts: The artifacts required for the SQLAlchemy model source code.

    Returns:
        The SQLAlchemy model source code.

    """
    template = jinja2.Template(_SQLALCHEMY_TEMPLATE)

    arg_init_source = arg_init(artifacts=artifacts.arg)
    arg_from_dict_source = arg_from_dict(artifacts=artifacts.arg)

    return template.render(
        artifacts=artifacts,
        arg_init_source=arg_init_source,
        arg_from_dict_source=arg_from_dict_source,
    )


def typed_dict_required(*, artifacts: types.TypedDictArtifacts) -> str:
    """
    Generate the TypedDict for required properties source code.

    Args:
        artifacts: The artifacts required for the TypedDict source code.

    Returns:
        The TypedDict for required properties source code.

    """
    template = jinja2.Template(_TYPED_DICT_REQUIRED_TEMPLATE)
    return template.render(artifacts=artifacts)


def typed_dict_not_required(*, artifacts: types.TypedDictArtifacts) -> str:
    """
    Generate the TypedDict for not required properties source code.

    Args:
        artifacts: The artifacts required for the TypedDict source code.

    Returns:
        The TypedDict for not required properties source code.

    """
    template = jinja2.Template(_TYPED_DICT_NOT_REQUIRED_TEMPLATE)
    return template.render(artifacts=artifacts)


def _arg_single_required(artifacts: types.ColumnArgArtifacts, name: str) -> str:
    """
    Transform the name and type of a single required argument to the input source.

    Args:
        artifacts: The artifacts for generating the argument for a column.
        name: The attribute name to use for the type.

    Returns:
        The source for the argument for the column.

    """
    return f", {artifacts.name}: {getattr(artifacts, name)}"


def _arg_single_not_required(artifacts: types.ColumnArgArtifacts, name: str) -> str:
    """
    Transform the name and type of a single not required argument to the input source.

    Args:
        artifacts: The artifacts for generating the argument for a column.
        name: The attribute name to use for the type.

    Returns:
        The source for the argument for the column.

    """
    required_source = _arg_single_required(artifacts, name)
    return f"{required_source} = {artifacts.default}"


def _arg(*, artifacts: types.ArgArtifacts, name: str) -> str:
    """
    Generate the arguments for a function signature of a model.

    Args:
        artifacts: The artifacts for the arguments.
        name: The attribute name to use for the type.

    Returns:
        The argument signature for the functions.

    """
    required_sources = map(
        lambda artifacts: _arg_single_required(artifacts, name), artifacts.required
    )
    not_required_sources = map(
        lambda artifacts: _arg_single_not_required(artifacts, name),
        artifacts.not_required,
    )
    return f'{"".join(required_sources)}{"".join(not_required_sources)}'


def arg_init(*, artifacts: types.ArgArtifacts) -> str:
    """
    Generate the arguments for the signature of __init__ for a model.

    Args:
        artifacts: The artifacts for the arguments.

    Returns:
        The argument signature for the __init__ functions.

    """
    return _arg(artifacts=artifacts, name="init_type")


def _check_not_read_only_arg(arg: types.ColumnArgArtifacts) -> bool:
    """Check whether an argument is not read only."""
    return arg.read_only is not True


def remove_read_only_args(*, artifacts: types.ArgArtifacts) -> types.ArgArtifacts:
    """
    Remove read only arguments from the artifacts.

    Args:
        artifacts: The artifacts to filter.

    Returns:
        The filtered artifacts.

    """
    return types.ArgArtifacts(
        required=list(filter(_check_not_read_only_arg, artifacts.required)),
        not_required=list(filter(_check_not_read_only_arg, artifacts.not_required)),
    )


def arg_from_dict(*, artifacts: types.ArgArtifacts) -> str:
    """
    Generate the arguments for the signature of from_dict for a model.

    Args:
        artifacts: The artifacts for the arguments.

    Returns:
        The argument signature for the from_dict functions.

    """
    no_read_only_artifacts = remove_read_only_args(artifacts=artifacts)
    return _arg(artifacts=no_read_only_artifacts, name="from_dict_type")


def generate(*, artifacts: types.ModelArtifacts) -> str:
    """
    Generate the overall template with the TypedDict and SQLAlchemy model.

    Args:
        artifacts: The artifacts for the model.

    Returns:
        The source code for the template.

    """
    # Construct individual source code
    sqlalchemy_source = sqlalchemy(artifacts=artifacts.sqlalchemy)
    typed_dict_required_source = typed_dict_required(artifacts=artifacts.typed_dict)
    typed_dict_not_required_source = typed_dict_not_required(
        artifacts=artifacts.typed_dict
    )

    # Construct overall source code
    template = jinja2.Template(_TEMPLATE, trim_blocks=True)
    return template.render(
        artifacts=artifacts,
        typed_dict_required=typed_dict_required_source,
        typed_dict_not_required=typed_dict_not_required_source,
        sqlalchemy=sqlalchemy_source,
    )



"""Base class providing utilities for SQLAlchemy models."""

import functools
import json
import typing

from .. import exceptions
from .. import facades
from .. import helpers
from .. import types as oa_types
from . import from_dict
from . import repr_
from . import to_dict

TUtilityBase = typing.TypeVar("TUtilityBase", bound="UtilityBase")
TOptUtilityBase = typing.Optional[TUtilityBase]


class UtilityBase:
    """Base class providing utilities for SQLAlchemy models."""

    # Record of the schema used to construct the model. Must be an object type. For all
    # columns any $ref must be resolved an allOf must be merged for all. Objects must
    # be recorded as a free-form object and have a x-de-$ref extension property with
    # the de-referenced name of the schema.
    _schema: typing.ClassVar[oa_types.Schema]

    def __init__(self, **kwargs: typing.Any) -> None:
        """Construct."""
        raise NotImplementedError

    @classmethod
    def _get_schema(cls) -> oa_types.Schema:
        """
        Get the schema.

        Raise ModelAttributeError if _schema is not defined.

        Returns:
            The schema.

        """
        # Checking for _schema
        if not hasattr(cls, "_schema"):
            raise exceptions.ModelAttributeError(
                "Model does not have a record of its schema. "
                "To support to_dict set the _schema class variable."
            )
        return cls._schema

    @classmethod
    def get_properties(cls) -> oa_types.Schema:
        """
        Get the properties from the schema.

        Raise ModelAttributeError if _schema is not defined.
        Raise MalformedSchemaError if the schema does not have any properties.

        Returns:
            The properties of the schema.

        """
        schema = cls._get_schema()
        # Checking that _schema has properties
        properties = schema.get("properties")
        if properties is None:
            raise exceptions.MalformedSchemaError(
                "The model schema does not have any properties.", schema=schema
            )
        return properties

    @staticmethod
    def _get_parent(*, schema: oa_types.Schema) -> typing.Type[TUtilityBase]:
        """Get the parent model of a model."""
        parent_name = helpers.ext_prop.get(source=schema, name="x-inherits")
        if parent_name is None or not isinstance(parent_name, str):
            raise exceptions.MalformedSchemaError(
                "To construct a model that inherits x-inherits must be present and a "
                "string.",
                schema=schema,
                x_inherits=parent_name,
                x_inherits_type=type(parent_name),
            )
        # Try to get model
        parent: TOptUtilityBase = facades.models.get_model(name=parent_name)
        if parent is None:
            raise exceptions.SchemaNotFoundError(
                "The parent model was not found on open_alchemy.models.",
                schema=schema,
                parent_model_name=parent_name,
            )
        return parent

    @classmethod
    def construct_from_dict_init(
        cls: typing.Type[TUtilityBase], **kwargs: typing.Any
    ) -> typing.Dict[str, typing.Any]:
        """Construct the dictionary passed to model construction."""
        # Check dictionary
        schema = cls._get_schema()
        try:
            facades.jsonschema.validate(instance=kwargs, schema=schema)
        except facades.jsonschema.ValidationError as exc:
            raise exceptions.MalformedModelDictionaryError(
                "The dictionary passed to from_dict is not a valid instance of the "
                "model schema.",
                schema=schema,
                kwargs=kwargs,
            ) from exc

        # Assemble dictionary for construction
        properties = cls.get_properties()
        model_dict: typing.Dict[str, typing.Any] = {}
        for name, value in kwargs.items():
            # Get the specification and type of the property
            property_schema = properties.get(name)
            if property_schema is None:
                raise exceptions.MalformedModelDictionaryError(
                    "A parameter was passed in that is not a property in the model "
                    "schema.",
                    parameter_name=name,
                    schema=schema,
                )

            # Convert to column value
            try:
                model_dict[name] = from_dict.convert(
                    value=value, schema=property_schema
                )
            except exceptions.BaseError as exc:
                exc.schema = schema  # type: ignore
                exc.property_schema = property_schema  # type: ignore
                exc.property_name = name  # type: ignore
                exc.property_value = value  # type: ignore
                raise

        return model_dict

    @classmethod
    def from_dict(cls: typing.Type[TUtilityBase], **kwargs: typing.Any) -> TUtilityBase:
        """
        Construct model instance from a dictionary.

        Raise MalformedModelDictionaryError when the dictionary does not satisfy the
        model schema.

        Args:
            kwargs: The values to construct the class with.

        Returns:
            An instance of the model constructed using the dictionary.

        """
        schema = cls._get_schema()
        # Handle model that inherits
        if helpers.schema.inherits(schema=schema, schemas={}):
            # Retrieve parent model
            parent: typing.Type[UtilityBase] = cls._get_parent(schema=schema)

            # Construct parent initialization dictionary
            # Get properties for schema
            properties = cls.get_properties()
            # Pass kwargs that don't belong to the current model to the parent
            parent_kwargs = {
                key: value for key, value in kwargs.items() if key not in properties
            }
            parent_init_dict = parent.construct_from_dict_init(**parent_kwargs)

            # COnstruct child (the current model) initialization dictionary
            child_kwargs = {
                key: value for key, value in kwargs.items() if key in properties
            }
            init_dict = {
                **parent_init_dict,
                **cls.construct_from_dict_init(**child_kwargs),
            }
        else:
            init_dict = cls.construct_from_dict_init(**kwargs)

        return cls(**init_dict)

    @classmethod
    def from_str(cls: typing.Type[TUtilityBase], value: str) -> TUtilityBase:
        """
        Construct model instance from a JSON string.

        Raise MalformedModelDictionaryError when the value is not a string or the string
        is not valid JSON.

        Args:
            kwargs: The values to construct the class with.

        Returns:
            An instance of the model constructed using the dictionary.

        """
        if not isinstance(value, str):
            raise exceptions.MalformedModelDictionaryError(
                "The value is not of type string.", value=value, value_type=type(value)
            )
        try:
            dict_value = json.loads(value)
        except json.JSONDecodeError as exc:
            raise exceptions.MalformedModelDictionaryError(
                "The string value is not valid JSON.", value=value
            ) from exc
        if not isinstance(dict_value, dict):
            raise exceptions.MalformedModelDictionaryError(
                "The string value is not a Python dictionary.",
                value=value,
                value_type=type(value),
            )
        return cls.from_dict(**dict_value)

    @classmethod
    def instance_to_dict(cls, instance: TUtilityBase) -> typing.Dict[str, typing.Any]:
        """Convert instance of the model to a dictionary."""
        schema = cls._get_schema()
        properties = cls.get_properties()

        # Collecting the values of the properties
        return_dict: typing.Dict[str, typing.Any] = {}
        for name, property_schema in properties.items():
            # Handle for writeOnly
            if helpers.peek.write_only(schema=property_schema, schemas={}):
                continue

            value = getattr(instance, name, None)

            # Handle none value
            if value is None:
                return_none = to_dict.return_none(schema=schema, property_name=name)
                if return_none is True:
                    return_dict[name] = None
                # Don't consider for coverage due to coverage bug
                continue  # pragma: no cover

            try:
                return_dict[name] = to_dict.convert(schema=property_schema, value=value)
            except exceptions.BaseError as exc:
                exc.schema = schema  # type: ignore
                exc.property_schema = property_schema  # type: ignore
                exc.property_name = name  # type: ignore
                exc.property_value = value  # type: ignore
                raise

        return return_dict

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        """
        Convert model instance to dictionary.

        Returns:
            The dictionary representation of the model.

        """
        schema = self._get_schema()
        if helpers.schema.inherits(schema=schema, schemas={}):
            # Retrieve parent model and convert to dict
            parent: typing.Type[UtilityBase] = self._get_parent(schema=schema)
            parent_dict = parent.instance_to_dict(self)
            return {**parent_dict, **self.instance_to_dict(self)}

        return self.instance_to_dict(self)

    def to_str(self) -> str:
        """
        Convert model instance to a string.

        Returns:
            The JSON string representation of the model.

        """
        instance_dict = self.to_dict()
        return json.dumps(instance_dict)

    __str__ = to_str

    def __repr__(self) -> str:
        """Calculate the repr for the model."""
        properties = self.get_properties()
        return repr_.calculate(instance=self, properties=properties)



"""Calculate the model from the schema artifacts."""

import sys
import typing

from open_alchemy import schemas

from .. import types
from . import args as _args
from . import column as _column
from . import type_ as _type
from . import typed_dict as _typed_dict


def calculate(
    *, artifacts: schemas.artifacts.types.ModelArtifacts, name: str
) -> types.ModelArtifacts:
    """
    Calculate the model artifacts from the schema.

    Args:
        schema: The schema of the model
        name: The name of the model.

    Returns:
        The artifacts for the model.

    """
    columns = list(_column.calculate(artifacts=artifacts))
    typed_dict_required_props, typed_dict_not_required_props = _typed_dict.calculate(
        artifacts=artifacts
    )
    required_args, not_required_args = _args.calculate(artifacts=artifacts)

    # Calculate model parent class
    parent_cls: str
    if sys.version_info[1] < 8:
        parent_cls = "typing_extensions.Protocol"
    else:  # version compatibility
        parent_cls = "typing.Protocol"

    # Calculate whether property lists are empty, their names and parent class
    typed_dict_required_empty = not typed_dict_required_props
    typed_dict_not_required_empty = not typed_dict_not_required_props
    typed_dict_required_name = None
    typed_dict_not_required_name: typing.Optional[str] = f"{name}Dict"
    typed_dict_required_parent_class = None
    typed_dict_not_required_parent_class: typing.Optional[str]
    if sys.version_info[1] < 8:
        typed_dict_not_required_parent_class = "typing_extensions.TypedDict"
    else:  # version compatibility
        typed_dict_not_required_parent_class = "typing.TypedDict"
    if not typed_dict_required_empty and not typed_dict_not_required_empty:
        typed_dict_required_parent_class = typed_dict_not_required_parent_class
        typed_dict_required_name = f"_{name}DictBase"
        typed_dict_not_required_parent_class = typed_dict_required_name
    if not typed_dict_required_empty and typed_dict_not_required_empty:
        typed_dict_required_name = typed_dict_not_required_name
        typed_dict_not_required_name = None
        typed_dict_required_parent_class = typed_dict_not_required_parent_class
        typed_dict_not_required_parent_class = None

    return types.ModelArtifacts(
        sqlalchemy=types.SQLAlchemyModelArtifacts(
            name=name,
            columns=columns,
            empty=not columns,
            arg=types.ArgArtifacts(
                required=required_args, not_required=not_required_args
            ),
            parent_cls=parent_cls,
            description=artifacts.description,
        ),
        typed_dict=types.TypedDictArtifacts(
            required=types.TypedDictClassArtifacts(
                props=typed_dict_required_props,
                empty=typed_dict_required_empty,
                name=typed_dict_required_name,
                parent_class=typed_dict_required_parent_class,
            ),
            not_required=types.TypedDictClassArtifacts(
                props=typed_dict_not_required_props,
                empty=typed_dict_not_required_empty,
                name=typed_dict_not_required_name,
                parent_class=typed_dict_not_required_parent_class,
            ),
        ),
    )



"""Facades for ringfenced modules."""
# pylint: disable=useless-import-alias

from . import code_formatter as code_formatter
from . import jsonschema as jsonschema
from . import models as models
from . import sqlalchemy as sqlalchemy



"""Validate the full schema (property, source and referenced)."""

from ..... import exceptions
from ..... import helpers as oa_helpers
from ..... import types as oa_types
from .... import helpers
from ... import helpers as validation_helpers
from ... import model
from ... import types
from .. import simple


def _check_pre_defined_property_schema(
    *,
    property_name: str,
    property_schema: oa_types.Schema,
    schema: oa_types.Schema,
    schemas: oa_types.Schemas,
    foreign_key: str,
):
    """
    Check for a pre-defined property on a schema.

    Assume property_schema has already been checked for validity.

    Args:
        property_name: The expected foreign key property name to check for.
        property_schema: The schema for the foreign key.
        schema: The schema to check for the property on.
        schemas: Used to resolve any $ref.
        foreign_key: The foreign key value.

    Returns:
        A result if something is wrong with the reason or None otherwise.

    """
    # Get the pre-defined property schema if it exists
    properties = helpers.iterate.properties_items(
        schema=schema, schemas=schemas, stay_within_tablename=True
    )
    filtered_properties = filter(lambda arg: arg[0] == property_name, properties)
    defined_property = next(filtered_properties, None)
    if defined_property is None:
        return None

    # Validate the schema
    _, defined_property_schema = defined_property
    schema_result = simple.check(schemas, defined_property_schema)
    if not schema_result.valid:
        return types.Result(
            False,
            f"{property_name} property :: {schema_result.reason}",
        )

    # Check that key information matches
    checks = (
        ("type", oa_helpers.peek.type_),
        ("format", oa_helpers.peek.format_),
        ("maxLength", oa_helpers.peek.max_length),
        ("default", oa_helpers.peek.default),
    )
    for key, func in checks:
        match_result = validation_helpers.value.check_matches(
            func=func,
            reference_schema=property_schema,
            check_schema=defined_property_schema,
            schemas=schemas,
        )
        if match_result is None:
            continue

        return types.Result(
            valid=False, reason=f"{property_name} :: {key} :: {match_result}"
        )

    # Check the foreign key
    actual_foreign_key = oa_helpers.peek.foreign_key(
        schema=defined_property_schema, schemas=schemas
    )
    if actual_foreign_key is None:
        return types.Result(
            False,
            f"{property_name} must define a foreign key",
        )
    if actual_foreign_key != foreign_key:
        return types.Result(
            False,
            f"the x-foreign-key of {property_name} is wrong, expected {foreign_key}, "
            f"the actual is {actual_foreign_key}",
        )

    return None


def _check_target_schema(
    *,
    target_schema: oa_types.Schema,
    schemas: oa_types.Schemas,
    column_name: str,
    modify_schema: oa_types.Schema,
    foreign_key_property_name: str,
) -> types.Result:
    """
    Check the schema that is targeted by a foreign key.

    Assume target_schema is valid.

    Args:
        target_schema: The schema targeted by a foreign key.
        schemas: The schemas used to resolve any $ref.
        column_name: The name of the foreign key column.
        modify_schema: The schema to add the foreign key property to.
        foreign_key_property_name: The name of the foreign key property to define.

    Returns:
        A result if something is wrong with the reason or None otherwise.

    """
    # Look for foreign key property schema
    properties = helpers.iterate.properties_items(
        schema=target_schema,
        schemas=schemas,
        stay_within_tablename=True,
    )
    filtered_properties = filter(lambda arg: arg[0] == column_name, properties)
    foreign_key_target_property = next(filtered_properties, None)
    if foreign_key_target_property is None:
        return types.Result(
            False,
            f"foreign key targeted schema must have the {column_name} property",
        )

    # Validate the schema
    (
        foreign_key_target_property_name,
        foreign_key_target_property_schema,
    ) = foreign_key_target_property
    schema_result = simple.check(schemas, foreign_key_target_property_schema)
    if not schema_result.valid:
        return types.Result(
            False,
            f"{foreign_key_target_property_name} property :: {schema_result.reason}",
        )

    # Check for pre-defined foreign key property
    foreign_key = oa_helpers.foreign_key.calculate_foreign_key(
        column_name=column_name,
        target_schema=target_schema,
        schemas=schemas,
    )
    pre_defined_result = _check_pre_defined_property_schema(
        property_name=foreign_key_property_name,
        property_schema=foreign_key_target_property_schema,
        schema=modify_schema,
        schemas=schemas,
        foreign_key=foreign_key,
    )
    if pre_defined_result is not None:
        return pre_defined_result

    return types.Result(True, None)


def _check_many_to_many_schema(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """
    Check one of the many to many schemas.

    Args:
        schema: The schema to check.
        schemas: Used to resolve any $ref.

    Returns:
        A result of the schema is not valid with a reason or None.

    """
    model_result = model.check(schema=schema, schemas=schemas)
    if not model_result.valid:
        return model_result

    # Check for primary key
    properties = helpers.iterate.properties_items(
        schema=schema, schemas=schemas, stay_within_tablename=True
    )
    primary_key_properties = filter(
        lambda args: oa_helpers.peek.primary_key(schema=args[1], schemas=schemas)
        is True,
        properties,
    )
    primary_key_property = next(primary_key_properties, None)
    if primary_key_property is None:
        return types.Result(False, "schema must have a primary key")

    # Check for multiple primary keys
    next_primary_key_property = next(primary_key_properties, None)
    if next_primary_key_property is not None:
        return types.Result(
            False,
            "many-to-many relationships currently only support single primary key "
            "schemas",
        )

    # Check property schema
    primary_key_property_name, primary_key_property_schema = primary_key_property
    schema_result = simple.check(schemas, primary_key_property_schema)
    if schema_result.valid is False:
        return types.Result(
            False, f"{primary_key_property_name} property :: {schema_result.reason}"
        )

    return None


def _check_many_to_many(
    *,
    parent_schema: oa_types.Schema,
    property_schema: oa_types.Schema,
    schemas: oa_types.Schemas,
) -> types.Result:
    """
    Check many-to-many relationships.

    Args:
        parent_schema: The schema that has the property that defines the relationship.
        property_schema: The schema of the items for the property that defines the
            relationship.
        schemas: Used to resolve any $ref.

    Returns:
        Whether the relationship is valid and the reason if it is not.

    """
    # Checking source schema
    source_result = _check_many_to_many_schema(schema=parent_schema, schemas=schemas)
    if source_result is not None:
        return types.Result(
            source_result.valid, f"source schema :: {source_result.reason}"
        )

    # Checking referenced schema
    _, ref_schema = oa_helpers.relationship.get_ref_schema_many_to_x(
        property_schema=property_schema, schemas=schemas
    )
    ref_result = _check_many_to_many_schema(schema=ref_schema, schemas=schemas)
    if ref_result is not None:
        return types.Result(
            ref_result.valid, f"referenced schema :: {ref_result.reason}"
        )

    return types.Result(True, None)


def _check_backref_property_properties_basics(
    property_name: str,
    backref_schema: oa_types.Schema,
    schemas: oa_types.Schemas,
) -> types.OptResult:
    """
    Check the backref schema.

    Args:
        parent_schema: The schema that has the property embedded in it.
        property_name: The name of the property.
        backref_schema: The schema of the back reference.
        schemas: All defined schemas used to resolve any $ref.
    """
    # Check for object type
    type_ = oa_helpers.peek.type_(schema=backref_schema, schemas=schemas)
    if type_ != "object":
        return types.Result(False, "the back reference schema must be an object")

    # Check properties values
    properties_values_result = validation_helpers.properties.check_properties_values(
        schema=backref_schema, schemas=schemas
    )
    if properties_values_result is not None:
        return properties_values_result

    # Check whether any property names match the property name
    properties_items = helpers.iterate.properties_items(
        schema=backref_schema, schemas=schemas
    )
    property_name_matches = next(
        filter(lambda args: args[0] == property_name, properties_items), None
    )
    if property_name_matches is not None:
        return types.Result(
            False,
            "properties cannot contain the property name of the relartionship to avoid "
            "circular references",
        )

    return None


def _check_backref_property_properties(
    parent_schema: oa_types.Schema,
    property_name: str,
    backref_schema: oa_types.Schema,
    schemas: oa_types.Schemas,
) -> types.OptResult:
    """
    Check the backref schema.

    Args:
        parent_schema: The schema that has the property embedded in it.
        property_name: The name of the property.
        backref_schema: The schema of the back reference.
        schemas: All defined schemas used to resolve any $ref.
    """
    basics_result = _check_backref_property_properties_basics(
        property_name=property_name, backref_schema=backref_schema, schemas=schemas
    )
    if basics_result is not None:
        return basics_result

    # Check for backreference properties not in the parent schema properties
    parent_properties_items = helpers.iterate.properties_items(
        schema=parent_schema, schemas=schemas
    )
    parent_properties = dict(parent_properties_items)
    properties_items = helpers.iterate.properties_items(
        schema=backref_schema, schemas=schemas
    )
    property_name_not_in_parent = next(
        filter(lambda args: args[0] not in parent_properties, properties_items), None
    )
    if property_name_not_in_parent is not None:
        name, _ = property_name_not_in_parent
        return types.Result(False, f"could not find {name} in the model schema")

    # Check properties are dictionaries
    properties_items_result = validation_helpers.properties.check_properties_items(
        schema=backref_schema, schemas=schemas
    )
    if properties_items_result is not None:
        return properties_items_result

    # Check schema matches
    checks = (
        ("type", oa_helpers.peek.type_),
        ("format", oa_helpers.peek.format_),
        ("maxLength", oa_helpers.peek.max_length),
        ("default", oa_helpers.peek.default),
    )
    for key, func in checks:
        properties_items = helpers.iterate.properties_items(
            schema=backref_schema, schemas=schemas
        )
        # pylint: disable=cell-var-from-loop
        # Calculate result for each property
        properties_items_value_results = map(
            lambda args: (
                args[0],
                validation_helpers.value.check_matches(
                    func=func,
                    reference_schema=parent_properties[args[0]],
                    check_schema=args[1],
                    schemas=schemas,
                ),
            ),
            properties_items,
        )
        # Look for the first failed result
        properties_items_value_result = next(
            filter(lambda args: args[1] is not None, properties_items_value_results),
            None,
        )
        if properties_items_value_result is not None:
            property_name, result = properties_items_value_result
            assert result is not None
            return types.Result(
                valid=False, reason=f"{property_name} :: {key} :: {result}"
            )

    return None


_BACKREF_EXPECTED_TYPE = {
    oa_types.RelationshipType.MANY_TO_ONE: "array",
    oa_types.RelationshipType.ONE_TO_ONE: "object",
    oa_types.RelationshipType.ONE_TO_MANY: "object",
    oa_types.RelationshipType.MANY_TO_MANY: "array",
}


def _check_backref_property(
    parent_schema: oa_types.Schema,
    property_name: str,
    relationship_type: oa_types.RelationshipType,
    backref_property_schema: oa_types.Schema,
    schemas: oa_types.Schemas,
) -> types.OptResult:
    """Check the back reference property."""
    # Check the type of the property
    read_only = oa_helpers.peek.read_only(
        schema=backref_property_schema, schemas=schemas
    )
    if not read_only:
        return types.Result(False, "the property must be readOnly")

    # Check the type of the backref property
    property_type = oa_helpers.peek.type_(
        schema=backref_property_schema, schemas=schemas
    )
    expected_property_type = _BACKREF_EXPECTED_TYPE[relationship_type]
    if expected_property_type != property_type:
        return types.Result(
            False,
            f"unexpected type, expected {expected_property_type} actual "
            f"{property_type}",
        )

    # Check the properties
    if relationship_type in {
        oa_types.RelationshipType.MANY_TO_ONE,
        oa_types.RelationshipType.MANY_TO_MANY,
    }:
        items_schema = oa_helpers.peek.items(
            schema=backref_property_schema, schemas=schemas
        )
        if items_schema is None:
            return types.Result(False, "items must be defined")
        properties_result = _check_backref_property_properties(
            parent_schema=parent_schema,
            property_name=property_name,
            backref_schema=items_schema,
            schemas=schemas,
        )
        if properties_result is not None:
            return types.Result(
                False, f"items :: properties :: {properties_result.reason}"
            )
    else:
        properties_result = _check_backref_property_properties(
            parent_schema=parent_schema,
            property_name=property_name,
            backref_schema=backref_property_schema,
            schemas=schemas,
        )
        if properties_result is not None:
            return types.Result(False, f"properties :: {properties_result.reason}")
    return None


def _check_backref(
    parent_schema: oa_types.Schema,
    property_name: str,
    relationship_type: oa_types.RelationshipType,
    property_schema: oa_types.Schema,
    schemas: oa_types.Schemas,
) -> types.OptResult:
    """
    Check the back reference, if defined.

    Assume the property schema and parent schema is valid.

    Algorithm:
        1. check for back reference, if not there stop,
        2. check for whether back reference value is in the properties, if not stop,
        3. check the type of the property based on the relationship type,
        4. if the back reference is an array, retrieve the items schema
        5. check whether the properties are valid,
        6. check that the property being processed is not in the properties,
        6. check that all defined properties are also on the parent schema and
        7. check that the type and format match.

    Args:
        parent_schema: The schema that has the property embedded in it.
        property_name: The name of the property.
        relationship_type: The type of the relationship the property defines.
        property_schema: The schema of the property.
        schemas: All defined schemas used to resolve any $ref.

    Returns:
        If the back reference is invalid, returns result with reason, otherwise None.

    """
    backref = helpers.backref.get(schema=property_schema, schemas=schemas)
    if backref is None:
        return None

    # Get the object schema
    object_schema: oa_types.Schema
    if relationship_type in {
        oa_types.RelationshipType.ONE_TO_MANY,
        oa_types.RelationshipType.MANY_TO_MANY,
    }:
        items_schema = oa_helpers.peek.items(schema=property_schema, schemas=schemas)
        assert items_schema is not None
        object_schema = items_schema
    else:
        object_schema = property_schema

    # Look for backref property
    properties = helpers.iterate.properties_items(schema=object_schema, schemas=schemas)
    property_ = next(filter(lambda args: args[0] == backref, properties), None)
    if property_ is None:
        return None
    _, backref_property_schema = property_

    try:
        return _check_backref_property(
            parent_schema=parent_schema,
            property_name=property_name,
            relationship_type=relationship_type,
            backref_property_schema=backref_property_schema,
            schemas=schemas,
        )

    except (exceptions.MalformedSchemaError, exceptions.TypeMissingError) as exc:
        return types.Result(False, f"malformed schema :: {exc}")


def check(
    schemas: oa_types.Schemas,
    parent_schema: oa_types.Schema,
    property_name: str,
    property_schema: oa_types.Schema,
) -> types.Result:
    """
    Check the source, referenced and property schema.

    Assume the property schema validation is already complete.

    At a high level:
    1. either an id or the column configured by x-foreign-key-column must be on the
        schema the foreign key points at,
    2. the property the foreign key points at must
        a. have a valid type and not object nor array and
        b. must have valid values for format, nullable, maxLength and default if they
            are defined
    3. the schema the foreign key points at must define a tablename except for
        many-to-many relationships,
    4. if a property on the schema the foreign key points at with the same name as the
        foreign key is already defined then it must match the following information:
        a. type,
        b. format,
        c. maxLength,
        d. default and
        e. have the same foreign key constraint as defined by the relationship,
    5. for many-to-many relationships both the source and referenced schema must have
        have a single primary key column with a valid type that is not an object nor an
        array and
    6. for many-to-many relationships both the source and referenced schema must have a
        single valid primary key property.

    Args:
        schemas: All defined schemas used to resolve any $ref.
        parent_schema: The schema that has the property embedded in it.
        property_name: The name of the property.
        property_schema: The schema of the property.

    Returns:
        WHether the full relationship schema is valid and, if not, why it isn't.

    """
    try:
        type_ = oa_helpers.relationship.calculate_type(
            schema=property_schema, schemas=schemas
        )

        if type_ in {
            oa_types.RelationshipType.MANY_TO_ONE,
            oa_types.RelationshipType.ONE_TO_ONE,
            oa_types.RelationshipType.ONE_TO_MANY,
        }:
            # Retrieve information required to check x-to-one and one-to-many
            # relationships
            column_name = oa_helpers.foreign_key.calculate_column_name(
                type_=type_,
                property_schema=property_schema,
                schemas=schemas,
            )
            target_schema = oa_helpers.foreign_key.get_target_schema(
                type_=type_,
                parent_schema=parent_schema,
                property_schema=property_schema,
                schemas=schemas,
            )
            target_schema_model_result = model.check(
                schema=target_schema, schemas=schemas
            )
            if not target_schema_model_result.valid:
                return types.Result(
                    False,
                    f"foreign key target schema :: {target_schema_model_result.reason}",
                )
            modify_schema = oa_helpers.foreign_key.get_modify_schema(
                type_=type_,
                parent_schema=parent_schema,
                property_schema=property_schema,
                schemas=schemas,
            )
            foreign_key_property_name = oa_helpers.foreign_key.calculate_prop_name(
                type_=type_,
                column_name=column_name,
                property_name=property_name,
                target_schema=target_schema,
                schemas=schemas,
            )
            result = _check_target_schema(
                target_schema=target_schema,
                schemas=schemas,
                column_name=column_name,
                modify_schema=modify_schema,
                foreign_key_property_name=foreign_key_property_name,
            )
            if not result.valid:
                return result
        else:
            result = _check_many_to_many(
                parent_schema=parent_schema,
                property_schema=property_schema,
                schemas=schemas,
            )
            if not result.valid:
                return result

        backref_result = _check_backref(
            parent_schema=parent_schema,
            property_name=property_name,
            relationship_type=type_,
            property_schema=property_schema,
            schemas=schemas,
        )
        if backref_result is not None:
            return types.Result(
                backref_result.valid, f"backref property :: {backref_result.reason}"
            )

        return types.Result(True, None)

    except exceptions.MalformedSchemaError as exc:
        return types.Result(False, f"malformed schema :: {exc}")



"""Validate the schema of a model."""

import typing

from ... import exceptions
from ... import helpers as oa_helpers
from ... import table_args
from ... import types as oa_types
from .. import helpers
from . import helpers as validation_helpers
from . import types


def _check_properties(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check properties."""
    # Check property values
    properties_values_result = validation_helpers.properties.check_properties_values(
        schema=schema, schemas=schemas
    )
    if properties_values_result is not None:
        return properties_values_result

    # Check there is at least a single property
    properties_items = helpers.iterate.properties_items(
        schema=schema, schemas=schemas, stay_within_model=True
    )
    first_property = next(properties_items, None)
    if first_property is None:
        return types.Result(False, "models must have at least 1 property themself")

    # Check that all property names are strings
    properties_items_result = validation_helpers.properties.check_properties_items(
        schema=schema, schemas=schemas
    )
    if properties_items_result is not None:
        return properties_items_result

    return None


def _get_property_names_model(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> typing.Iterator[str]:
    """Retrieve all property names."""
    properties_items = helpers.iterate.properties_items(
        schema=schema, schemas=schemas, stay_within_model=True
    )
    return map(lambda prop: prop[0], properties_items)


def _check_required(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check required."""
    # Retrieve property names
    property_name_set = set(_get_property_names_model(schema=schema, schemas=schemas))

    # Check that all required values are lists
    required_values = helpers.iterate.required_values(schema=schema, schemas=schemas)
    any_required_value_not_list = any(
        filter(
            lambda required_value: not isinstance(required_value, list), required_values
        )
    )
    if any_required_value_not_list:
        return types.Result(False, "value of required must be a list")

    # Check required values
    required_items = helpers.iterate.required_items(
        schema=schema, schemas=schemas, stay_within_model=True
    )
    required_items_set = set(required_items)
    any_required_items_not_string = any(
        filter(
            lambda required_item: not isinstance(required_item, str), required_items_set
        )
    )
    if any_required_items_not_string:
        return types.Result(False, "required :: all items must be strings")

    # Check that all required items are property names
    required_not_properties = required_items_set - property_name_set
    if required_not_properties:
        return types.Result(
            False,
            "required :: all items must be properties, "
            f"{next(iter(required_not_properties))} is not",
        )

    return None


def _check_mandatory(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check mandatory keys of the model schema."""
    # Check for inheritance
    inherits = oa_helpers.schema.inherits(schema=schema, schemas=schemas)
    if inherits:
        oa_helpers.inheritance.get_parent(schema=schema, schemas=schemas)

    tablename = oa_helpers.peek.tablename(schema=schema, schemas=schemas)
    if tablename is None and not inherits:
        return types.Result(False, "every model must define x-tablename")

    type_ = oa_helpers.peek.type_(schema=schema, schemas=schemas)
    if type_ != "object":
        return types.Result(False, "models must have the object type")

    properties_result = _check_properties(schema=schema, schemas=schemas)
    if properties_result is not None:
        return types.Result(
            properties_result.valid, f"properties :: {properties_result.reason}"
        )

    return None


def _check_kwargs(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check kwargs value of it exists."""
    kwargs = oa_helpers.peek.kwargs(schema=schema, schemas=schemas)
    if kwargs is None:
        return None

    def invalid_key(key) -> bool:
        """Check value of key is valid."""
        assert isinstance(key, str)
        return not key.startswith("__") or not key.endswith("__")

    kwargs_keys = list(kwargs.keys())
    any_kwargs_keys_invalid = any(
        filter(
            invalid_key,
            kwargs_keys,
        )
    )
    if any_kwargs_keys_invalid:
        return types.Result(
            False, "models x-kwargs must have keys that start and end with __"
        )
    if "__table_args__" in kwargs:
        return types.Result(False, "models x-kwargs cannot define __table_args__")

    return None


def _check_invalid_keys(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check for keys that are not valid on a model."""
    invalid_keys = (
        "x-primary-key",
        "x-autoincrement",
        "x-index",
        "x-unique",
        "x-foreign-key",
        "x-foreign-key-kwargs",
    )
    seen_invalid_keys = filter(
        lambda key: oa_helpers.peek.peek_key(schema=schema, schemas=schemas, key=key)
        is not None,
        invalid_keys,
    )
    invalid_key = next(seen_invalid_keys, None)
    if invalid_key is not None:
        return types.Result(False, f"models do not support the {invalid_key} key")

    return None


def _get_property_names_table(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> typing.Iterator[str]:
    """Retrieve all property names."""
    properties_items = helpers.iterate.properties_items(
        schema=schema, schemas=schemas, stay_within_tablename=True
    )
    return map(lambda prop: prop[0], properties_items)


def _check_modifiers(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check basics related to tablename, inheritance and type."""
    required_result = _check_required(schema=schema, schemas=schemas)
    if required_result is not None:
        return required_result

    # Check description value
    oa_helpers.peek.description(schema=schema, schemas=schemas)
    # Check mixins value
    oa_helpers.peek.mixins(schema=schema, schemas=schemas)

    # Check kwargs
    kwargs_result = _check_kwargs(schema=schema, schemas=schemas)
    if kwargs_result is not None:
        return kwargs_result

    # Check for invalid keys
    invalid_keys_result = _check_invalid_keys(schema=schema, schemas=schemas)
    if invalid_keys_result is not None:
        return invalid_keys_result

    # Check composite index
    index_spec = oa_helpers.peek.composite_index(schema=schema, schemas=schemas)
    if index_spec is not None:
        index_expressions = set(
            table_args.factory.iter_index_expressions(spec=index_spec)
        )
        property_names = set(_get_property_names_table(schema=schema, schemas=schemas))
        index_not_properties = index_expressions - property_names
        if index_not_properties:
            return types.Result(
                False,
                "x-composite-index :: all expressions must be properties, "
                f"{next(iter(index_not_properties))} is not",
            )

    # Check composite unique
    unique_spec = oa_helpers.peek.composite_unique(schema=schema, schemas=schemas)
    if unique_spec is not None:
        unique_columns = set(table_args.factory.iter_unique_columns(spec=unique_spec))
        property_names = set(_get_property_names_table(schema=schema, schemas=schemas))
        unique_not_properties = unique_columns - property_names
        if unique_not_properties:
            return types.Result(
                False,
                "x-composite-unique :: all columns must be properties, "
                f"{next(iter(unique_not_properties))} is not",
            )

    return None


def check(schemas: oa_types.Schemas, schema: oa_types.Schema) -> types.Result:
    """
    Check that a schema is valid.

    Args:
        schemas: All defined schemas used to resolve any $ref.
        schema: The schema to validate.

    Returns:
        Whether the schema is valid with the reason if it is not.

    """
    try:
        mandatory_result = _check_mandatory(schema=schema, schemas=schemas)
        if mandatory_result is not None:
            return mandatory_result
        modifiers_result = _check_modifiers(schema=schema, schemas=schemas)
        if modifiers_result is not None:
            return modifiers_result

    except (exceptions.MalformedSchemaError, exceptions.TypeMissingError) as exc:
        return types.Result(False, f"malformed schema :: {exc}")
    except exceptions.SchemaNotFoundError as exc:
        return types.Result(False, f"reference :: {exc}")
    except exceptions.MalformedExtensionPropertyError as exc:
        return types.Result(False, f"extension property :: {exc}")

    return types.Result(True, None)



"""Helpers for schemas."""

import typing

from .. import exceptions
from .. import types
from . import all_of as all_of_helper
from . import peek
from . import ref as ref_helper


def constructable(*, schema: types.Schema, schemas: types.Schemas) -> bool:
    """
    Check that a schema is constructable.

    The rules are:
    1. the resolved schema must have either x-tablename or x-inherits and
    2. the schema cannot just be a local $ref.

    Args:
        schema: The schema to check.

    Returns:
        Whether the schema is constructable.

    """
    if not isinstance(schema, dict):
        return False
    # Check for reference only models
    ref = schema.get("$ref")
    if ref is not None and (not isinstance(ref, str) or ref.startswith("#")):
        return False
    # Check allOf
    all_of = schema.get("allOf")
    if all_of is not None and (not isinstance(all_of, list) or len(all_of) < 2):
        if not isinstance(all_of, list) or not all_of:
            return False
        return constructable(schema=all_of[0], schemas=schemas)
    # Check for tablename and inherits
    tablename = peek.peek_key(schema=schema, schemas=schemas, key="x-tablename")
    try:
        schema_inherits = inherits(schema=schema, schemas=schemas)
    except exceptions.MalformedSchemaError:
        return True
    return tablename is not None or schema_inherits is True


def inherits(*, schema: types.Schema, schemas: types.Schemas) -> typing.Optional[bool]:
    """
    Check whether a schema inherits.

    Args:
        schema: The schema to check.
        schemas: All the schemas.

    Returns:
        Whether the schema inherits.

    """
    inherits_value = peek.inherits(schema=schema, schemas=schemas)
    if inherits_value is None:
        return None
    if isinstance(inherits_value, str) or inherits_value is True:
        return True
    return False


def prepare(
    *,
    schema: types.Schema,
    schemas: types.Schemas,
    skip_name: typing.Optional[str] = None,
) -> types.Schema:
    """
    Resolve $ref and merge allOf.

    Args:
        schema: The schema to prepare.
        schemas: The schemas from which to resolve any $ref.
        skip_name (optional): Any schema name to skip.

    Returns:
        The prepared schema.

    """
    _, schema = ref_helper.resolve(
        name="", schema=schema, schemas=schemas, skip_name=skip_name
    )
    return all_of_helper.merge(schema=schema, schemas=schemas, skip_name=skip_name)


def prepare_deep(schema: types.Schema, schemas: types.Schemas):
    """
    Resolve $ref and merge allOf including for object properties and items.

    Assume the schema is a valid JSONSchema.

    Args:
        schema: The schema to prepare.
        schemas: The schemas from which to resolve any $ref.

    Returns:
        The prepared schema.

    """
    schema = prepare(schema=schema, schemas=schemas)

    # Resolve $ref in any properties
    properties = schema.get("properties", None)
    if properties is not None:
        for name, prop_schema in properties.items():
            properties[name] = prepare_deep(schema=prop_schema, schemas=schemas)

    # Resolve $ref of any items
    items_schema = schema.get("items", None)
    if items_schema is not None:
        schema["items"] = prepare_deep(schema=items_schema, schemas=schemas)

    return schema



"""Retrieve model artifacts."""

import typing

from ... import helpers as oa_helpers
from ... import table_args
from ... import types as oa_types
from .. import helpers
from . import types


def _calculate_backref(schema: oa_types.Schema) -> types.ModelBackrefArtifacts:
    """Calculate the backref artifact from the backref schema."""
    type_ = oa_helpers.peek.type_(schema=schema, schemas={})
    assert type_ in {"object", "array"}

    if type_ == "object":
        parent = oa_helpers.ext_prop.get(source=schema, name="x-de-$ref")
        assert isinstance(parent, str)
        return types.ModelBackrefArtifacts(types.BackrefSubType.OBJECT, parent)

    items_schema = oa_helpers.peek.items(schema=schema, schemas={})
    assert items_schema is not None
    parent = oa_helpers.ext_prop.get(source=items_schema, name="x-de-$ref")
    assert isinstance(parent, str)
    return types.ModelBackrefArtifacts(types.BackrefSubType.ARRAY, parent)


def get(
    schemas: oa_types.Schemas, schema: oa_types.Schema
) -> types.ModelExPropertiesArtifacts:
    """
    Retrieve the artifacts for the model.

    Assume that the schema is valid.

    Args:
        schema: The model schema.
        schemas: All the defined schemas used to resolve any $ref.

    Returns:
        The artifacts for the model.

    """
    tablename = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.tablename, schema=schema, schemas=schemas
    )
    assert tablename is not None
    inherits = oa_helpers.schema.inherits(schema=schema, schemas=schemas)
    parent: typing.Optional[str] = None
    if inherits is True:
        parent = oa_helpers.inheritance.get_parent(schema=schema, schemas=schemas)

    description = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.description, schema=schema, schemas=schemas
    )

    mixins = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.mixins, schema=schema, schemas=schemas
    )

    kwargs = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.kwargs, schema=schema, schemas=schemas
    )

    composite_index: typing.Optional[oa_types.IndexList] = None
    composite_index_value = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.composite_index, schema=schema, schemas=schemas
    )
    if composite_index_value is not None:
        composite_index = table_args.factory.map_index(spec=composite_index_value)

    composite_unique: typing.Optional[oa_types.UniqueList] = None
    composite_unique_value = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.composite_unique, schema=schema, schemas=schemas
    )
    if composite_unique_value is not None:
        composite_unique = table_args.factory.map_unique(spec=composite_unique_value)

    backrefs = helpers.iterate.backrefs_items(schema=schema, schemas=schemas)
    backrefs_artifacts = map(
        lambda args: (args[0], _calculate_backref(args[1])), backrefs
    )

    return types.ModelExPropertiesArtifacts(
        tablename=tablename,
        inherits=inherits,
        parent=parent,
        description=description,
        mixins=mixins,
        kwargs=kwargs,
        composite_index=composite_index,
        composite_unique=composite_unique,
        backrefs=list(backrefs_artifacts),
    )



"""Retrieve artifacts for a JSON property."""

import copy

from .... import helpers as oa_helpers
from .... import types as oa_types
from ... import helpers
from .. import types


def get(
    schemas: oa_types.Schemas, schema: oa_types.Schema, required: bool = False
) -> types.JsonPropertyArtifacts:
    """
    Retrieve the artifacts for a JSON property.

    Args:
        schemas: All the defined schemas.
        schema: The schema of the JSON property to gather artifacts for.
        required: WHether the property appears in the required list.

    Returns:
        The artifacts for the property.

    """
    schema = copy.deepcopy(
        oa_helpers.schema.prepare_deep(schema=schema, schemas=schemas)
    )

    nullable = oa_helpers.peek.nullable(schema=schema, schemas=schemas)

    description = oa_helpers.peek.description(schema=schema, schemas=schemas)

    read_only = oa_helpers.peek.read_only(schema=schema, schemas=schemas)
    write_only = oa_helpers.peek.write_only(schema=schema, schemas=schemas)

    primary_key = oa_helpers.peek.primary_key(schema=schema, schemas=schemas)
    index = oa_helpers.peek.index(schema=schema, schemas=schemas)
    unique = oa_helpers.peek.unique(schema=schema, schemas=schemas)

    foreign_key = oa_helpers.peek.foreign_key(schema=schema, schemas=schemas)

    kwargs = oa_helpers.peek.kwargs(schema=schema, schemas=schemas)
    foreign_key_kwargs = oa_helpers.peek.foreign_key_kwargs(
        schema=schema, schemas=schemas
    )

    # Remove extension properties from schema
    helpers.clean.extension(schema=schema)
    # Add in x-json
    schema["x-json"] = True

    return types.JsonPropertyArtifacts(
        type=oa_types.PropertyType.JSON,
        description=description,
        schema=schema,
        required=required,
        open_api=types.OpenApiJsonPropertyArtifacts(
            nullable=nullable,
            read_only=read_only,
            write_only=write_only,
        ),
        extension=types.ExtensionJsonPropertyArtifacts(
            primary_key=primary_key is True,
            index=index,
            unique=unique,
            foreign_key=foreign_key,
            kwargs=kwargs,
            foreign_key_kwargs=foreign_key_kwargs,
        ),
    )



"""Apply automatic code formatting to source code."""


def apply(*, source: str) -> str:
    """
    Apply automatic code formatting to source code.

    Args:
        source: The source code.

    Returns:
        The formatted source code.

    """
    return source



"""Calculate the typed dict artifacts for a model."""

import typing

from open_alchemy import schemas
from open_alchemy import types as oa_types

from .. import types
from . import type_


class ReturnValue(typing.NamedTuple):
    """The return value for the calculated artifacts."""

    required: typing.List[types.ColumnArtifacts]
    not_required: typing.List[types.ColumnArtifacts]


def _get_write_only(
    artifacts: schemas.artifacts.types.TAnyPropertyArtifacts,
) -> typing.Optional[bool]:
    """Get write only for any property artifacts."""
    if artifacts.type == oa_types.PropertyType.SIMPLE:
        return artifacts.open_api.write_only
    if artifacts.type == oa_types.PropertyType.JSON:
        return artifacts.open_api.write_only
    if artifacts.type == oa_types.PropertyType.RELATIONSHIP:
        return artifacts.write_only
    return None


def _calculate(
    *,
    artifacts: typing.Iterable[
        typing.Tuple[str, schemas.artifacts.types.TAnyPropertyArtifacts]
    ]
) -> typing.Iterable[types.ColumnArtifacts]:
    """Calculate the typed dict artifacts from property artifacts."""
    no_dict_ignore_properties = filter(
        lambda args: not (
            args[1].type == oa_types.PropertyType.SIMPLE
            and args[1].extension.dict_ignore
        ),
        artifacts,
    )
    no_dict_ignore_no_write_only = filter(
        lambda args: not _get_write_only(args[1]),
        no_dict_ignore_properties,
    )
    return map(
        lambda args: types.ColumnArtifacts(
            name=args[0],
            type=type_.typed_dict(artifacts=args[1]),
            description=args[1].description,
        ),
        no_dict_ignore_no_write_only,
    )


def calculate(*, artifacts: schemas.artifacts.types.ModelArtifacts) -> ReturnValue:
    """
    Calculate the typed dict artifacts from model schema artifacts.

    Args:
        artifacts: The schema artifacts for a model.

    Returns:
        The artifacts for the typed dict.

    """
    required = filter(lambda args: args[1].required, artifacts.properties)
    not_required = filter(lambda args: not args[1].required, artifacts.properties)

    return ReturnValue(
        required=list(_calculate(artifacts=required)),
        not_required=list(_calculate(artifacts=not_required)),
    )



"""Calculate the repr for the model."""

import typing

from open_alchemy import types


def calculate(*, instance: typing.Any, properties: types.Schema) -> str:
    """
    Calculate the repr for the model.

    The repr is the string that would be needed to create an equivalent instance of the
    model.

    Args:
        instance: The model instance to calculate the repr for.
        properties: The properties of the model instance.

    Returns:
        The string that would be needed to create an equivalent instance of the model.

    """
    # Calculate the name
    name = type(instance).__name__

    # Retrieve property values
    prop_repr_gen = (
        (prop, repr(getattr(instance, prop, None))) for prop in properties.keys()
    )
    prop_repr_str_gen = (f"{name}={value}" for name, value in prop_repr_gen)
    prop_repr_str = ", ".join(prop_repr_str_gen)

    # Calculate repr
    return f"open_alchemy.models.{name}({prop_repr_str})"



"""Convert an OpenAPI value based on the format to the equivalent Python value."""

import datetime
import typing

from open_alchemy import exceptions
from open_alchemy import types


def convert(
    *, value: types.TColumnDefault, type_: str, format_: typing.Optional[str]
) -> types.TPyColumnDefault:
    """
    Convert an OpenAPI value to it's python type based on the format.

    Args:
        value: The value to convert.
        type_: The OpenAPI type.
        format_: The OpenAPI format.

    Returns:
        The value converted to its equivalent Python type.

    """
    if type_ in {"object", "array"}:
        raise exceptions.MalformedSchemaError(
            "Cannot convert object nor array types to Python equivalent values."
        )

    if isinstance(value, str) and format_ == "date":
        try:
            return datetime.date.fromisoformat(value)
        except ValueError as exc:
            raise exceptions.MalformedSchemaError("Invalid date string.") from exc
    if isinstance(value, str) and format_ == "date-time":
        try:
            return datetime.datetime.fromisoformat(value)
        except ValueError as exc:
            raise exceptions.MalformedSchemaError("Invalid date-time string.") from exc
    if isinstance(value, str) and format_ == "binary":
        return value.encode()
    if format_ == "double":
        raise exceptions.MalformedSchemaError("Double format is not supported.")
    return value



"""Helpers for back references."""

import typing

from ... import helpers
from ... import types


def get(schemas: types.Schemas, schema: types.Schema) -> typing.Optional[str]:
    """
    Get the back reference from the property schema.

    The following rules are used:
    1. if there is an items key, recursively call on the items value.
    1. peek for x-backrefs on the schema and return if found.
    3. Return None.

    Args:
        schemas: All the defined schemas.
        schema: The schema of the property.

    Returns:
        Whether the property defines a back reference.

    """
    # Handle items
    items_schema = helpers.peek.items(schema=schema, schemas=schemas)
    if items_schema is not None:
        return get(schema=items_schema, schemas=schemas)

    # Peek for backref
    return helpers.peek.backref(schema=schema, schemas=schemas)


def defined(schemas: types.Schemas, schema: types.Schema) -> bool:
    """
    Check whether the property schema defines a back reference.

    The following rules are used:
    1. if there is an items key, recursively call on the items value.
    1. peek for x-backrefs on the schema and return True if found.
    3. Return False.

    Args:
        schemas: All the defined schemas.
        schema: The schema of the property.

    Returns:
        Whether the property defines a back reference.

    """
    return get(schemas, schema) is not None



"""Functions to expose iterables for schemas."""

import functools
import typing

from ... import exceptions
from ... import helpers
from ... import types


def constructable(
    *, schemas: types.Schemas
) -> typing.Iterator[typing.Tuple[str, types.Schema]]:
    """
    Create an iterable with all constructable schemas from all schemas.

    Iterates over all items in the schemas, checks whether a schema is constructable and
    yields those that are.

    Args:
        schemas: The schemas to iterate over.

    Returns:
        iterable with all schemas that are constructable.

    """
    for name, schema in schemas.items():
        try:
            if not helpers.schema.constructable(schema=schema, schemas=schemas):
                continue
        except (exceptions.MalformedSchemaError, exceptions.SchemaNotFoundError):
            continue

        yield name, schema


def not_constructable(
    *, schemas: types.Schemas
) -> typing.Iterator[typing.Tuple[str, types.Schema]]:
    """
    Create an iterable with all non-constructable schemas from all schemas.

    Iterates over all items in the schemas, checks whether a schema is not constructable
    and yields those that are.

    Args:
        schemas: The schemas to iterate over.

    Returns:
        iterable with all schemas that are not constructable.

    """
    for name, schema in schemas.items():
        try:
            if helpers.schema.constructable(schema=schema, schemas=schemas):
                continue
        except (exceptions.MalformedSchemaError, exceptions.SchemaNotFoundError):
            continue

        yield name, schema


def _calculate_skip_name(
    *,
    schema: types.Schema,
    schemas: types.Schemas,
    stay_within_tablename: bool = False,
    stay_within_model: bool = False,
) -> typing.Optional[str]:
    """
    Calculate the references to skip depending on the inheritance.

    Args:
        schema: The constructable schems.
        schemas: All defined schemas (not just the constructable ones).
        stay_within_tablename: Ensures that on properties on the same table are
            iterated over. For joined table inheritance, the reference to the parent is
            not followed.
        stay_within_model: Ensures that each property is only returned once. For both
            single and joined table inheritance no reference to the parent is followed.

    Returns:
        The references to skip or None.

    """
    if not stay_within_tablename and not stay_within_model:
        return None

    inheritance_type = helpers.inheritance.calculate_type(
        schema=schema, schemas=schemas
    )
    if inheritance_type != helpers.inheritance.Type.NONE:
        parent_name = helpers.inheritance.retrieve_parent(
            schema=schema, schemas=schemas
        )

        # Check for single
        if stay_within_model:
            return parent_name

        # Check for JOINED
        if (
            not stay_within_model
            and inheritance_type == helpers.inheritance.Type.JOINED_TABLE
            and stay_within_tablename
        ):
            return parent_name

    return None


def _filter_duplicates(seen_keys: typing.Set[str], args) -> bool:
    """Remove duplicate values."""
    key, _ = args

    if key in seen_keys:
        return False

    seen_keys.add(key)
    return True


def properties_items(
    *,
    schema: types.Schema,
    schemas: types.Schemas,
    stay_within_tablename: bool = False,
    stay_within_model: bool = False,
) -> typing.Iterator[typing.Any]:
    """
    Create an iterable with all properties of a schema from a constructable schema.

    Checks for $ref, if it is there resolves to the underlying schema and recursively
    processes that schema.
    Checks for allOf, if it is there recursively processes each schema.
    Otherwise looks for properties and yields all items if the key exists.

    Args:
        schema: The constructable schems.
        schemas: All defined schemas (not just the constructable ones).
        stay_within_tablename: Ensures that only properties on the same table are
            iterated over. For joined table inheritance, the reference to the parent is
            not followed.
        stay_within_model: Ensures that each property is only returned once. For both
            single and joined table inheritance no reference to the parent is followed.

    Returns:
        An iterator with all properties of a schema.

    """
    init_filter_duplicates = functools.partial(_filter_duplicates, set())

    properties_values_iterator = properties_values(
        schema=schema,
        schemas=schemas,
        stay_within_tablename=stay_within_tablename,
        stay_within_model=stay_within_model,
    )
    for properties_value in properties_values_iterator:
        if not isinstance(properties_value, dict):
            continue
        yield from filter(init_filter_duplicates, properties_value.items())


def properties_values(
    *,
    schema: types.Schema,
    schemas: types.Schemas,
    stay_within_tablename: bool = False,
    stay_within_model: bool = False,
) -> typing.Iterator[typing.Any]:
    """
    Return iterable with all values of the properties key of the constructable schema.

    Checks for $ref, if it is there resolves to the underlying schema and recursively
    processes that schema.
    Checks for allOf, if it is there recursively processes each schema.
    Otherwise yields the properties key value.

    Args:
        schema: The constructable schems.
        schemas: All defined schemas (not just the constructable ones).
        stay_within_tablename: Ensures that only properties on the same table are
            iterated over. For joined table inheritance, the reference to the parent is
            not followed.
        stay_within_model: Ensures that each properties value is only returned once. For
            both single and joined table inheritance no reference to the parent is
            followed.

    Returns:
        An iterator with all properties key values.

    """
    skip_name: typing.Optional[str] = None
    try:
        skip_name = _calculate_skip_name(
            schema=schema,
            schemas=schemas,
            stay_within_tablename=stay_within_tablename,
            stay_within_model=stay_within_model,
        )
    except (
        exceptions.MalformedSchemaError,
        exceptions.InheritanceError,
        exceptions.SchemaNotFoundError,
    ):
        return

    yield from _any_key(
        schema=schema, schemas=schemas, skip_name=skip_name, key="properties"
    )


def _any_key(
    *,
    schema: types.Schema,
    schemas: types.Schemas,
    skip_name: typing.Optional[str],
    key: str,
) -> typing.Iterator[typing.Any]:
    """Private interface for properties."""
    if not isinstance(schema, dict):
        return

    # Handle $ref
    if schema.get("$ref") is not None:
        try:
            _, ref_schema = helpers.ref.resolve(
                name="", schema=schema, schemas=schemas, skip_name=skip_name
            )
        except (exceptions.MalformedSchemaError, exceptions.SchemaNotFoundError):
            return
        yield from _any_key(
            schema=ref_schema, schemas=schemas, skip_name=skip_name, key=key
        )

    # Handle allOf
    all_of = schema.get("allOf")
    if all_of is not None:
        if not isinstance(all_of, list):
            return

        all_of_dicts = list(
            filter(lambda sub_schema: isinstance(sub_schema, dict), all_of)
        )

        # Process not $ref first
        all_of_no_ref = filter(
            lambda sub_schema: not helpers.peek.peek_key(
                schema=sub_schema, schemas=schemas, key="$ref"
            ),
            all_of_dicts,
        )
        for sub_schema in all_of_no_ref:
            yield from _any_key(
                schema=sub_schema, schemas=schemas, skip_name=skip_name, key=key
            )

        # Process $ref
        all_of_ref = filter(
            lambda sub_schema: helpers.peek.peek_key(
                schema=sub_schema, schemas=schemas, key="$ref"
            ),
            all_of_dicts,
        )
        for sub_schema in all_of_ref:
            yield from _any_key(
                schema=sub_schema, schemas=schemas, skip_name=skip_name, key=key
            )

    # Handle simple case
    value = schema.get(key)
    if value is None:
        return
    yield value


def required_values(
    *,
    schema: types.Schema,
    schemas: types.Schemas,
    stay_within_model: bool = False,
) -> typing.Iterator[typing.Any]:
    """
    Return iterable with all values of the required key of the constructable schema.

    Checks for $ref, if it is there resolves to the underlying schema and recursively
    processes that schema.
    Checks for allOf, if it is there recursively processes each schema.
    Otherwise yields the required key value.

    Args:
        schema: The constructable schems.
        schemas: All defined schemas (not just the constructable ones).
        stay_within_model: Ensures that each required value is only returned once. For
            both single and joined table inheritance no reference to the parent is
            followed.

    Returns:
        An iterator with all required key values.

    """
    skip_name: typing.Optional[str] = None
    try:
        skip_name = _calculate_skip_name(
            schema=schema,
            schemas=schemas,
            stay_within_tablename=False,
            stay_within_model=stay_within_model,
        )
    except (
        exceptions.MalformedSchemaError,
        exceptions.InheritanceError,
        exceptions.SchemaNotFoundError,
    ):
        return

    yield from _any_key(
        schema=schema, schemas=schemas, skip_name=skip_name, key="required"
    )


def required_items(
    *,
    schema: types.Schema,
    schemas: types.Schemas,
    stay_within_model: bool = False,
) -> typing.Iterator[typing.Any]:
    """
    Return iterable with all items of the required key of the constructable schema.

    Checks for $ref, if it is there resolves to the underlying schema and recursively
    processes that schema.
    Checks for allOf, if it is there recursively processes each schema.
    Otherwise yields the required key value.

    Args:
        schema: The constructable schems.
        schemas: All defined schemas (not just the constructable ones).
        stay_within_model: Ensures that each required value is only returned once. For
            both single and joined table inheritance no reference to the parent is
            followed.

    Returns:
        An iterator with all items of the required key.

    """
    required_values_iterator = required_values(
        schema=schema, schemas=schemas, stay_within_model=stay_within_model
    )
    for required_value in required_values_iterator:
        if not isinstance(required_value, list):
            continue
        yield from required_value


def backrefs_items(
    *,
    schema: types.Schema,
    schemas: types.Schemas,
) -> typing.Iterator[typing.Any]:
    """
    Create an iterable with all backrefs of a schema from a constructable schema.

    Checks for $ref, if it is there resolves to the underlying schema and recursively
    processes that schema.
    Checks for allOf, if it is there recursively processes each schema.
    Otherwise looks for x-backrefs and yields all items if the key exists.

    Args:
        schema: The constructable schems.
        schemas: All defined schemas (not just the constructable ones).

    Returns:
        An iterator with all backrefs of a schema.

    """
    init_filter_duplicates = functools.partial(_filter_duplicates, set())

    backrefs_values_iterator = backrefs_values(
        schema=schema,
        schemas=schemas,
    )
    for backrefs_value in backrefs_values_iterator:
        if not isinstance(backrefs_value, dict):
            continue
        yield from filter(init_filter_duplicates, backrefs_value.items())


def backrefs_values(
    *,
    schema: types.Schema,
    schemas: types.Schemas,
) -> typing.Iterator[typing.Any]:
    """
    Return iterable with all values of the backrefs key of the constructable schema.

    Checks for $ref, if it is there resolves to the underlying schema and recursively
    processes that schema.
    Checks for allOf, if it is there recursively processes each schema.
    Otherwise yields the backrefs key value.

    Args:
        schema: The constructable schems.
        schemas: All defined schemas (not just the constructable ones).

    Returns:
        An iterator with all backrefs key values.

    """
    yield from _any_key(
        schema=schema, schemas=schemas, skip_name=None, key="x-backrefs"
    )



"""Define validation rules for simple properties."""

from .... import exceptions
from .... import helpers as oa_helpers
from .... import types as oa_types
from .. import types

TYPES = oa_helpers.type_.SIMPLE_TYPES

# Set of valid type and format combinations
_VALID_TYPE_FORMAT = {
    ("integer", None),
    ("integer", "int32"),
    ("integer", "int64"),
    ("number", None),
    ("number", "float"),
    ("boolean", None),
}


def _check_modifiers(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check property schema modifiers."""
    # check type
    type_ = oa_helpers.peek.type_(schema=schema, schemas=schemas)
    if type_ not in TYPES:
        return types.Result(False, f"{type_} type is not supported")

    # check format
    format_ = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.format_, schema=schema, schemas=schemas
    )
    type_format = (type_, format_)
    # Check type and format combination
    if type_ != "string" and type_format not in _VALID_TYPE_FORMAT:
        return types.Result(False, f"{format_} format is not supported for {type_}")

    # Define format display value
    format_str = ""
    if format_ is not None:
        format_str = f" {format_} format"

    # Check maxLength
    max_length = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.max_length, schema=schema, schemas=schemas
    )
    if max_length is not None:
        if type_ != "string" or format_ in {"date", "date-time"}:
            return types.Result(
                False, f"{type_}{format_str} does not support maxLength"
            )

    # Check autoincrement
    autoincrement = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.autoincrement, schema=schema, schemas=schemas
    )
    if autoincrement is not None and type_ != "integer":
        return types.Result(
            False, f"{type_}{format_str} does not support x-autoincrement"
        )

    return None


def check_kwargs(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check the value of kwargs."""
    # Check kwargs
    kwargs = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.kwargs, schema=schema, schemas=schemas
    )
    # Check for unexpected keys
    if kwargs is not None:
        unexpected_keys = {
            "nullable",
            "default",
            "server_default",
            "primary_key",
            "autoincrement",
            "index",
            "unique",
        }
        intersection = unexpected_keys.intersection(kwargs.keys())
        if intersection:
            return types.Result(
                False, f"x-kwargs :: may not contain the {next(iter(intersection))} key"
            )

    # Check foreign_key
    foreign_key = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.foreign_key, schema=schema, schemas=schemas
    )

    # Check foreign key kwargs
    foreign_key_kwargs = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.foreign_key_kwargs, schema=schema, schemas=schemas
    )
    if foreign_key_kwargs is not None and foreign_key is None:
        return types.Result(
            False, "x-foreign-key-kwargs :: can only be defined alongside x-foreign-key"
        )

    return None


def check(schemas: oa_types.Schemas, schema: oa_types.Schema) -> types.Result:
    """
    Check the schema of a simple property (not an object nor an array).

    Args:
        schemas: The schemas used to resolve any $ref.
        schema: The schema of the property.

    Returns:
        A result with whether the schema is valid and a reason if it is not.

    """
    try:
        # Check modifiers
        modifiers_result = _check_modifiers(schema=schema, schemas=schemas)
        if modifiers_result is not None:
            return modifiers_result

        # Check kwargs
        kwargs_result = check_kwargs(schema=schema, schemas=schemas)
        if kwargs_result is not None:
            return kwargs_result

        # Check nullable
        oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.nullable, schema=schema, schemas=schemas
        )
        # Check primary_key
        oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.primary_key, schema=schema, schemas=schemas
        )
        # Check index
        oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.index, schema=schema, schemas=schemas
        )
        # Check unique
        oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.unique, schema=schema, schemas=schemas
        )
        # Check description
        oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.description, schema=schema, schemas=schemas
        )
        # Check default
        oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.default, schema=schema, schemas=schemas
        )
        # Check server default
        oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.server_default, schema=schema, schemas=schemas
        )
        # Check readOnly
        oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.read_only, schema=schema, schemas=schemas
        )
        # Check writeOnly
        oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.write_only, schema=schema, schemas=schemas
        )

    except exceptions.SchemaNotFoundError as exc:
        return types.Result(False, f"reference :: {exc}")
    except (exceptions.MalformedSchemaError, exceptions.TypeMissingError) as exc:
        return types.Result(False, f"malformed schema :: {exc}")

    return types.Result(True, None)



"""Retrieve artifacts from the schemas."""

import typing

from ... import types as _oa_types
from .. import helpers as _helpers
from .. import validation
from . import model
from . import property_
from . import types


def _get_properties_artifacts(
    schemas: _oa_types.Schemas, schema: _oa_types.Schema
) -> typing.Iterable[typing.Tuple[str, types.TAnyPropertyArtifacts]]:
    """Get an iterator with properties artifacts."""
    # Get model properties
    properties = _helpers.iterate.properties_items(
        schema=schema, schemas=schemas, stay_within_model=True
    )
    required_set = set(_helpers.iterate.required_items(schema=schema, schemas=schemas))
    # Filter for valid properties
    valid_properties = filter(
        lambda args: validation.property_.check(
            schemas, schema, args[0], args[1]
        ).valid,
        properties,
    )
    return map(
        lambda args: (
            args[0],
            property_.get(schemas, schema, args[0], args[1], args[0] in required_set),
        ),
        valid_properties,
    )


def _get_properties(
    *, schemas: _oa_types.Schemas, schema: _oa_types.Schema
) -> types.PropertiesValue:
    """
    Get artifacts for the properties of a model.

    Args:
        schemas: All defined schemas used to resolve any $ref.
        schema_name: The name of the schema to validate.
        schema: The schema to validate.

    Returns:
        The artifacts for the properties.

    """
    properties_artifacts = _get_properties_artifacts(schemas, schema)
    properties_t_artifacts: typing.Iterable[
        typing.Tuple[str, types.PropertyValue]
    ] = map(
        lambda args: (args[0], {"artifacts": args[1].to_dict()}),
        properties_artifacts,
    )
    return dict(properties_t_artifacts)


def _get_model(
    schemas: _oa_types.Schemas, schema: _oa_types.Schema
) -> types.ModelValue:
    """
    Get artifacts for a model.

    Args:
        schemas: All defined schemas used to resolve any $ref.
        schema: The schema of the model to get artifacts for.

    Returns:
        The artifacts of the model.

    """
    model_artifacts = model.get(schemas, schema)
    return {
        "artifacts": model_artifacts.to_dict(),
        "properties": _get_properties(schemas=schemas, schema=schema),
    }


def get_models(*, schemas: _oa_types.Schemas) -> types.ModelsValue:
    """
    Get artifacts for the models of a schema.

    Assume the schemas is valid although any of its models may not.

    Args:
        schemas: The schemas to get artifacts for.

    Returns:
        The artifacts for each model.

    """
    constructables = _helpers.iterate.constructable(schemas=schemas)
    valid_constructables = filter(
        lambda args: validation.model.check(schemas, args[1]).valid, constructables
    )
    constructables_artifacts = map(
        lambda args: (args[0], _get_model(schemas, args[1])), valid_constructables
    )
    return dict(constructables_artifacts)


def get(*, spec: typing.Any) -> types.SpecValue:
    """
    Get artifacts for a specification.

    Args:
        spec: The specification to get artifacts from.

    Returns:
        The artifacts for the specification.

    """
    # Check spec to schemas
    spec_result = validation.spec_validation.check(spec=spec)
    if not spec_result.valid:
        return {}

    assert isinstance(spec, dict)
    components = spec.get("components")
    assert isinstance(components, dict)
    schemas = components.get("schemas")
    assert isinstance(schemas, dict)

    return {"models": get_models(schemas=schemas)}


def _from_schemas_get_properties_artifacts(
    stay_within_model: bool, schemas: _oa_types.Schemas, schema: _oa_types.Schema
) -> typing.Iterable[typing.Tuple[str, types.TAnyPropertyArtifacts]]:
    """Get an iterator with properties artifacts."""
    # Get model properties
    properties = _helpers.iterate.properties_items(
        schema=schema, schemas=schemas, stay_within_model=stay_within_model
    )
    required_set = set(_helpers.iterate.required_items(schema=schema, schemas=schemas))
    return map(
        lambda args: (
            args[0],
            property_.get(schemas, schema, args[0], args[1], args[0] in required_set),
        ),
        properties,
    )


def _from_schemas_get_model(
    stay_within_model: bool, schemas: _oa_types.Schemas, schema: _oa_types.Schema
) -> types.ModelArtifacts:
    """
    Get artifacts for a model.

    Assume the schema is valid.

    Args:
        schema: The schema of the model to get artifacts for.
        schemas: All defined schemas used to resolve any $ref.
        stay_within_model: Whether only properties from within a model should be
            included.

    Returns:
        The artifacts of the model.

    """
    model_artifacts = model.get(schemas, schema)
    properties_artifacts = _from_schemas_get_properties_artifacts(
        stay_within_model, schemas, schema
    )
    return types.ModelArtifacts(
        tablename=model_artifacts.tablename,
        inherits=model_artifacts.inherits,
        parent=model_artifacts.parent,
        description=model_artifacts.description,
        mixins=model_artifacts.mixins,
        kwargs=model_artifacts.kwargs,
        composite_index=model_artifacts.composite_index,
        composite_unique=model_artifacts.composite_unique,
        backrefs=model_artifacts.backrefs,
        properties=list(properties_artifacts),
    )


def get_from_schemas(
    *, schemas: _oa_types.Schemas, stay_within_model: bool
) -> types.ModelsModelArtifacts:
    """
    Get the artifacts from schemas.

    Assume the schemas are valid.

    Args:
        schemas: The schemas to get artifacts from.
        stay_within_model: Whether only properties from within a model should be
            included.

    Returns:
        The artifacts for the schemas.

    """
    constructables = _helpers.iterate.constructable(schemas=schemas)
    return dict(
        map(
            lambda args: (
                args[0],
                _from_schemas_get_model(stay_within_model, schemas, args[1]),
            ),
            constructables,
        )
    )



"""Construct column for relationship property."""

from .. import facades
from .. import types as oa_types
from . import types


def handle(
    *, artifacts: oa_types.TAnyRelationshipPropertyArtifacts
) -> types.TRelationship:
    """
    Handle a relationship property.

    Args:
        artifacts: The artifacts of the relationship property.

    Returns:
        The constructed column.

    """
    return facades.sqlalchemy.relationship.construct(artifacts=artifacts)



"""Validate the full schema (property, source and referenced)."""

import typing

from .... import helpers as oa_helpers
from .... import types as oa_types


def check_matches(
    *,
    func: oa_helpers.peek.PeekValue,
    reference_schema: oa_types.Schema,
    check_schema: oa_types.Schema,
    schemas: oa_types.Schemas,
) -> typing.Optional[str]:
    """
    Check that the value matches in two schemas.

    Args:
        func: Used to retrieve the value.
        reference_schema: The schema to check against.
        check_schema: The schema to check
        schemas: All defined schemas, used to resolve any $ref.

    Returns:
        An invalid result with reason if the values don't match otherwise None.

    """
    expected_value = oa_helpers.peek.prefer_local(
        get_value=func, schema=reference_schema, schemas=schemas
    )
    if expected_value is None:
        expected_value_str = "not to be defined"
    else:
        expected_value_str = f'"{expected_value}"'

    actual_value = oa_helpers.peek.prefer_local(
        get_value=func, schema=check_schema, schemas=schemas
    )
    if actual_value is None:
        actual_value_str = "not defined"
    else:
        actual_value_str = f'"{actual_value}"'

    if expected_value != actual_value:
        return f"expected {expected_value_str}, actual is {actual_value_str}"

    return None



"""Check that all x-tablename are unique."""

import typing

from ... import helpers as oa_helpers
from ... import types as oa_types
from .. import helpers
from . import types


def _is_single_table_inheritance(
    schemas: oa_types.Schemas, schema: oa_types.Schema
) -> bool:
    """
    Calculate whether the schema uses single table inheritance.

    Args:
        schemas: All defined schemas.
        schema: The schema to check.

    Returns:
        Whether the schema uses single table inheritance.

    """
    if not oa_helpers.schema.inherits(schema=schema, schemas=schemas):
        return False

    inheritance_type = oa_helpers.inheritance.calculate_type(
        schema=schema, schemas=schemas
    )
    return inheritance_type == oa_helpers.inheritance.Type.SINGLE_TABLE


def check(*, schemas: oa_types.Schemas) -> types.Result:
    """
    Check that all x-tablename are unique.

    Assume that schemas are otherwise valid.

    The algorithm is:
    1. iterate over constructable schemas,
    2. skip single table inheritance schemas,
    3. record the tablename and schema name,
    4. if a tablename has already been seen, return result that is not valid and
    5. return valid result.

    Args:
        schemas: The schemas to check.

    Returns:
        Whether the schemas have unique tablenames and the reason if not.

    """
    constructable_schemas = helpers.iterate.constructable(schemas=schemas)
    not_single_inheritance_schemas = filter(
        lambda args: not _is_single_table_inheritance(schemas, args[1]),
        constructable_schemas,
    )

    # Record the tablename against each schema name
    seen_tablenames: typing.Dict[str, str] = {}

    for name, schema in not_single_inheritance_schemas:
        # Retrieve tablename
        tablename = oa_helpers.peek.prefer_local(
            get_value=oa_helpers.peek.tablename, schemas=schemas, schema=schema
        )
        assert tablename is not None

        # Check whether the tablename has already been seen
        if tablename not in seen_tablenames:
            seen_tablenames[tablename] = name
            continue

        # Return result that is invalid with the reason
        seen_on_schema = seen_tablenames[tablename]
        return types.Result(
            valid=False,
            reason=(
                f'duplicate "x-tablename" value {tablename} defined on the schema '
                f"{name}, already defined on the schema {seen_on_schema}"
            ),
        )

    return types.Result(valid=True, reason=None)



"""Helpers for foreign keys."""

from .. import types
from . import peek
from . import ref as ref_helper
from . import relationship


def calculate_column_name(
    *,
    type_: types.RelationshipType,
    property_schema: types.Schema,
    schemas: types.Schemas,
) -> str:
    """
    Calculate the column name based on the relationship property schema.

    Assume the property is a x-to-one or one-to-many relationship.
    Assume the relationship property schema is valid.

    Args:
        type_: The type of relationship the schema defines.
        property_schema: The schema of the relationship property.
        schemas: All defined schemas used to resolve any $ref.

    Returns:
        The name of the foreign key column.

    """
    assert type_ != types.RelationshipType.MANY_TO_MANY

    if type_ == types.RelationshipType.ONE_TO_MANY:
        items_schema = peek.items(schema=property_schema, schemas=schemas)
        assert items_schema is not None
        property_schema = items_schema

    x_foreign_key_column = peek.prefer_local(
        get_value=peek.foreign_key_column, schema=property_schema, schemas=schemas
    )
    return x_foreign_key_column if x_foreign_key_column is not None else "id"


def get_target_schema(
    *,
    type_: types.RelationshipType,
    parent_schema: types.Schema,
    property_schema: types.Schema,
    schemas: types.Schemas,
) -> types.Schema:
    """
    Get the schema the foreign key targets.

    Assume type_ is not many-to-many.
    Assume property schema is a valid relationship.

    Args:
        type: The type of relationship.
        parent_schema: The schema the relationship property is embedded in.
        property_schema: The schema of the relationship property.
        schemas: Used to resolve any $ref.

    Returns:
        The schema that is the target of the foreign key.

    """
    assert type_ != types.RelationshipType.MANY_TO_MANY

    if type_ == types.RelationshipType.ONE_TO_MANY:
        return parent_schema

    ref = peek.ref(schema=property_schema, schemas=schemas)
    assert ref is not None
    _, ref_schema = ref_helper.get_ref(ref=ref, schemas=schemas)
    return ref_schema


def calculate_prop_name(
    *,
    type_: types.RelationshipType,
    column_name: str,
    property_name: str,
    target_schema: types.Schema,
    schemas: types.Schemas,
) -> str:
    """
    Calculate the foreign key property name based on the relationship type.

    Assume type_ is not a many to many relationship.
    Assume target_schema is a valid model.

    For x-to-one use the column and property name, for one-to-many also use the
    tablename.

    Args:
        type_: The type of relationship.
        column_name: The foreign key column name.
        property_name: The name of the property that defines the relationship.
        target_schema: The schema of the model targeted by the foreign key
            of the relationship.
        schema: All defines schemas used to resolve any $ref.

    Returns:
        The name of the foreign key property.

    """
    assert type_ != types.RelationshipType.MANY_TO_MANY

    if type_ == types.RelationshipType.ONE_TO_MANY:
        tablename = peek.prefer_local(
            get_value=peek.tablename, schema=target_schema, schemas=schemas
        )
        assert tablename is not None
        property_name = f"{tablename}_{property_name}"

    return f"{property_name}_{column_name}"


def calculate_foreign_key(
    *, column_name: str, target_schema: types.Schema, schemas: types.Schemas
) -> str:
    """
    Calculate the foreign key.

    Assume type_ is not a many to many relationship.
    Assume target_schema is a valid model.

    Args:
        column_name: The name of the foreign key column.
        target_schema: The schema of the model targeted by the foreign key
            of the relationship.
        schema: All defined schemas used to resolve any $ref.

    Returns:
        The foreign key.

    """
    tablename = peek.prefer_local(
        get_value=peek.tablename, schema=target_schema, schemas=schemas
    )
    assert tablename is not None
    return f"{tablename}.{column_name}"


def get_modify_schema(
    *,
    type_: types.RelationshipType,
    parent_schema: types.Schema,
    property_schema: types.Schema,
    schemas: types.Schemas,
) -> types.Schema:
    """
    Get the schema the foreign key needs to be added to.

    Assume type_ is not many-to-many.
    Assume property schema is a valid relationship.

    Args:
        type_: The type of relationship.
        parent_schema: The schema the relationship property is embedded in.
        property_schema: The schema of the relationship property.
        schemas: Used to resolve any $ref.

    Returns:
        The schema the foreign key needs to be defined on.

    """
    assert type_ != types.RelationshipType.MANY_TO_MANY

    if type_ == types.RelationshipType.ONE_TO_MANY:
        _, ref_schema = relationship.get_ref_schema_many_to_x(
            property_schema=property_schema, schemas=schemas
        )
        return ref_schema

    return parent_schema


def get_modify_name(
    *,
    type_: types.RelationshipType,
    parent_name: str,
    property_schema: types.Schema,
    schemas: types.Schemas,
) -> str:
    """
    Get the name of the schema the foreign key needs to be added to.

    Assume type_ is not many-to-many.
    Assume property schema is a valid relationship.

    Args:
        type: The type of relationship.
        parent_name: The name of the parent schema.
        property_schema: The schema of the relationship property.
        schemas: Used to resolve any $ref.

    Returns:
        The schema the foreign key needs to be defined on.

    """
    assert type_ != types.RelationshipType.MANY_TO_MANY

    if type_ == types.RelationshipType.ONE_TO_MANY:
        ref_name, _ = relationship.get_ref_schema_many_to_x(
            property_schema=property_schema, schemas=schemas
        )
        return ref_name

    return parent_name



"""Generate model from OpenAPI schema."""

import typing

from . import column_factory
from . import exceptions
from . import helpers
from . import mixins
from . import table_args
from . import types
from . import utility_base


def model_factory(
    *,
    name: str,
    get_base: types.GetBase,
    schemas: types.Schemas,
    artifacts: types.ModelsModelArtifacts,
) -> typing.Type:
    """
    Convert OpenAPI schema to SQLAlchemy model.

    Args:
        name: The name of the schema.
        get_base: Funcrtion to retrieve the base class for the model.
        schemas: The OpenAPI schemas.
        artifacts: The artifacts for the models.

    Returns:
        The model as a class.

    """
    schema = _get_schema(name, schemas)
    artifacts_name, _ = helpers.ref.resolve(
        name=name, schema=schemas[name], schemas=schemas
    )
    model_artifacts = artifacts.get(artifacts_name)
    assert model_artifacts is not None

    # Calculating the class variables for the model
    model_class_vars = {}
    required_exists = "required" in schema
    required_array = schema.get("required", [])
    # Initializing the schema to record for the model
    model_schema: types.Schema = {"type": "object", "properties": {}}
    if required_exists:
        model_schema["required"] = required_array
    if "x-inherits" in schema:
        model_schema["x-inherits"] = helpers.ext_prop.get(
            source=schema, name="x-inherits"
        )
    description = helpers.peek.description(schema=schema, schemas={})
    if description is not None:
        model_schema["description"] = description

    for prop_name, prop_artifacts in model_artifacts.properties:
        prop_column = column_factory.column_factory(artifacts=prop_artifacts)
        if prop_column is not None:
            model_class_vars[prop_name] = prop_column

        dict_ignore = (
            prop_artifacts.type == types.PropertyType.SIMPLE
            and prop_artifacts.extension.dict_ignore
        )

        if not dict_ignore:
            model_schema["properties"][prop_name] = prop_artifacts.schema

    # Retrieve mixins
    mixin_classes: typing.Tuple[typing.Type, ...] = tuple()
    mixin_values = helpers.peek.mixins(schema=schema, schemas=schemas)
    if mixin_values is not None:
        mixin_classes = mixins.get(mixins=mixin_values)

    # Assembling model
    base = get_base(name=name, schemas=schemas)
    return type(
        name,
        (base, utility_base.UtilityBase, *mixin_classes),
        {
            "_schema": model_schema,
            **model_class_vars,
            "__table_args__": table_args.construct(schema=schema),
            **_get_kwargs(schema=schema),
            **_prepare_model_dict(schema=schema),
            "__abstract__": False,
        },
    )


def _get_schema(name: str, schemas: types.Schemas) -> types.Schema:
    """
    Retrieve and prepare the schema from the schemas.

    Assume the schema has already been prepared. Replace x-inherits boolean with the
    parent name.

    Raise SchemaNotFoundError if the schema is not found in the schemas.
    Raise MalformedSchemaError is the schema does not have x-tablename.
    Raise FeatureNotImplementedError if the type is not found or it is not "object".
    Raise MalformedSchemaError if there are no properties.

    Args:
        name: The name of the schema to retrieve.
        schemas: All the schemas.

    Returns:
        The schema.

    """
    # Input validation
    # Checking that name is in schemas
    if name not in schemas:
        raise exceptions.SchemaNotFoundError(f"{name} not found in schemas")
    schema: types.Schema = schemas.get(name, {})
    # Check for parent
    inherits = helpers.schema.inherits(schema=schema, schemas=schemas)
    if not inherits:
        # De-referencing schema
        schema = helpers.schema.prepare(schema=schema, schemas=schemas)
        # Checking for tablename key
        if "x-tablename" not in schema:
            raise exceptions.MalformedSchemaError(
                f'"x-tablename" is a required schema property for {name}.'
            )
    else:
        parent = helpers.inheritance.retrieve_parent(schema=schema, schemas=schemas)
        # De-referencing schema excluding parent schema
        schema = helpers.schema.prepare(
            schema=schema, schemas=schemas, skip_name=parent
        )
        # Checking for inherits key
        if "x-inherits" not in schema:
            raise exceptions.MalformedSchemaError(
                f'"x-inherits" is a required schema property for {name}.'
            )
        schema["x-inherits"] = parent
    # Checking for object type
    if schema.get("type") != "object":
        raise exceptions.FeatureNotImplementedError(
            f"{schema.get('type')} is not supported in {name}."
        )
    if not schema.get("properties"):
        raise exceptions.MalformedSchemaError(
            f"At least 1 property is required for {name}."
        )
    return schema


def _get_kwargs(*, schema: types.Schema) -> types.TKwargs:
    """
    Retrieve kwargs for for the model.

    Raise MalformedExtensionPropertyError is the key does not start and end with __.

    Args:
        schema: The schema of the model.

    Returns:
        The kwargs for the model.

    """
    kwargs = helpers.ext_prop.get_kwargs(
        source=schema, reserved={"__tablename__", "__table_args__"}
    )
    if kwargs is None:
        return {}
    # Check that key starts and ends with __
    if any(not key.startswith("__") or not key.endswith("__") for key in kwargs.keys()):
        raise exceptions.MalformedExtensionPropertyError(
            "Model kwargs can only start and end with '__'."
        )
    return kwargs


def _prepare_model_dict(schema: types.Schema) -> typing.Dict[str, typing.Any]:
    """
    Prepare the dictionary used to construct the model.

    Assume the schema is valid.
    Add the __tablename__ key if it is required.

    Args:
        name: The name of the schema to prepare the dictionary for.
        schemas: All the schemas.

    Returns:
        The dictionary for the schema.

    """
    tablename = helpers.peek.tablename(schema=schema, schemas={})
    if tablename is not None:
        return {"__tablename__": tablename}
    return {}



"""Construct table args for a model."""

import itertools
import typing

from sqlalchemy import schema as sa_schema

from open_alchemy import helpers
from open_alchemy import types

from . import factory

TableArg = typing.Union[sa_schema.UniqueConstraint, sa_schema.Index]
TableArgs = typing.Tuple[TableArg, ...]


def construct(*, schema: types.Schema) -> TableArgs:
    """
    Construct any table args from the object schema.

    Look for x-composite-unique and x-composite-index keys in the schema and construct
    any unique constraints and indexes based on their value.

    Args:
        schema: The schema for the object.

    Returns:
        A tuple with any unique constraints and indexes.

    """
    # Keep track of any table arguments
    table_args: typing.List[typing.Iterable[TableArg]] = []

    # Handle x-composite-unique
    unique_spec = helpers.ext_prop.get(source=schema, name="x-composite-unique")
    if unique_spec is not None:
        table_args.append(factory.unique_factory(spec=unique_spec))
    # Handle x-composite-index
    index_spec = helpers.ext_prop.get(source=schema, name="x-composite-index")
    if index_spec is not None:
        table_args.append(factory.index_factory(spec=index_spec))

    return tuple(itertools.chain.from_iterable(table_args))



"""Validation for readOnly properties."""

from .... import exceptions
from .... import helpers as oa_helpers
from .... import types as oa_types
from ... import helpers
from .. import helpers as validation_helpers
from .. import types


def _check_object(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.Result:
    """Check readOnly object."""
    # Check properties values and items
    properties_values_result = validation_helpers.properties.check_properties_values(
        schema=schema, schemas=schemas
    )
    if properties_values_result is not None:
        return properties_values_result
    properties_items_result = validation_helpers.properties.check_properties_items(
        schema=schema, schemas=schemas
    )
    if properties_items_result is not None:
        return properties_items_result

    # Get types of properties
    properties_items = helpers.iterate.properties_items(schema=schema, schemas=schemas)
    properties_items_type = map(
        lambda args: (args[0], oa_helpers.peek.type_(schema=args[1], schemas=schemas)),
        properties_items,
    )
    properties_items_type = filter(
        lambda args: args[1] not in oa_helpers.type_.SIMPLE_TYPES, properties_items_type
    )
    first_properties_items_type = next(properties_items_type, None)
    if first_properties_items_type is not None:
        name, type_ = first_properties_items_type
        return types.Result(
            False,
            f"properties :: {name} :: readOnly object propeerties do not support the "
            f"{type_} type",
        )

    return types.Result(True, None)


def _check_array(*, schema: oa_types.Schema, schemas: oa_types.Schemas) -> types.Result:
    """Check readOnly array."""
    items_schema = oa_helpers.peek.items(schema=schema, schemas=schemas)
    if items_schema is None:
        return types.Result(False, "readOnly array properties must define items")

    # Check items type
    items_type = oa_helpers.peek.type_(schema=items_schema, schemas=schemas)
    if items_type != "object":
        return types.Result(
            False, "items :: readOnly array items must have the object type"
        )

    # Check items
    object_result = _check_object(schema=items_schema, schemas=schemas)
    if not object_result.valid:
        return types.Result(False, f"items :: {object_result.reason}")

    return types.Result(True, None)


def check(schemas: oa_types.Schemas, schema: oa_types.Schema) -> types.Result:
    """
    Check the schema of a readOnly property.

    Args:
        schemas: The schemas used to resolve any $ref.
        schema: The schema of the property.

    Returns:
        A result with whether the schema is valid and a reason if it is not.

    """
    try:
        type_ = oa_helpers.peek.type_(schema=schema, schemas=schemas)

        assert type_ not in oa_helpers.type_.SIMPLE_TYPES

        oa_helpers.peek.description(schema=schema, schemas=schemas)

        if type_ == "object":
            return _check_object(schema=schema, schemas=schemas)

        if type_ == "array":
            return _check_array(schema=schema, schemas=schemas)

        return types.Result(False, f"{type_} type is not supported")

    except (exceptions.MalformedSchemaError, exceptions.TypeMissingError) as exc:
        return types.Result(False, f"malformed schema :: {exc}")
    except exceptions.SchemaNotFoundError as exc:
        return types.Result(False, f"reference :: {exc}")



"""Retrieve artifacts for a relationship property."""

import typing

from .... import helpers as oa_helpers
from .... import types as oa_types
from .. import types


def _get_parent(*, schema: oa_types.Schema, schemas: oa_types.Schemas) -> str:
    """Retrieve the parent name from an object reference."""
    ref = oa_helpers.peek.ref(schema=schema, schemas=schemas)
    assert ref is not None
    parent, _ = oa_helpers.ref.get_ref(ref=ref, schemas=schemas)
    return parent


def _calculate_x_to_one_schema(
    *, parent: str, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> oa_types.ObjectRefSchema:
    """Calculate the schema for a x-to-one relationship."""
    return_schema: oa_types.ObjectRefSchema = {"type": "object", "x-de-$ref": parent}

    description = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.description, schema=schema, schemas=schemas
    )
    if description is not None:
        return_schema["description"] = description
    nullable = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.nullable, schema=schema, schemas=schemas
    )
    if nullable is not None:
        return_schema["nullable"] = nullable
    write_only = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.write_only, schema=schema, schemas=schemas
    )
    if write_only is not None:
        return_schema["writeOnly"] = write_only

    return return_schema


def _get_backref_property(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> typing.Optional[str]:
    """Retrieve the backref from an object reference."""
    return oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.backref, schema=schema, schemas=schemas
    )


def _get_kwargs(
    *, parent: str, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    """Retrieve the kwargs name from an object reference."""
    return oa_helpers.peek.peek_key(
        schema=schema, schemas=schemas, key="x-kwargs", skip_ref=parent
    )


def _get_write_only(
    *, parent: str, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> typing.Optional[bool]:
    """Retrieve the writeOnly value from a schema."""
    return oa_helpers.peek.peek_key(
        schema=schema, schemas=schemas, key="writeOnly", skip_ref=parent
    )


def _get_description(
    *, parent: str, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> typing.Optional[str]:
    """Retrieve the description value from a schema."""
    return oa_helpers.peek.peek_key(
        schema=schema, schemas=schemas, key="description", skip_ref=parent
    )


def _get_foreign_key(
    *,
    relationship_type: oa_types.RelationshipType,
    schema: oa_types.Schema,
    parent_schema: oa_types.Schema,
    schemas: oa_types.Schemas
) -> str:
    """Calculate the foreign key."""
    column_name = oa_helpers.foreign_key.calculate_column_name(
        type_=relationship_type,
        property_schema=schema,
        schemas=schemas,
    )
    target_schema = oa_helpers.foreign_key.get_target_schema(
        type_=relationship_type,
        parent_schema=parent_schema,
        property_schema=schema,
        schemas=schemas,
    )
    return oa_helpers.foreign_key.calculate_foreign_key(
        column_name=column_name,
        target_schema=target_schema,
        schemas=schemas,
    )


def _get_foreign_key_property(
    *,
    relationship_type: oa_types.RelationshipType,
    schema: oa_types.Schema,
    property_name: str,
    parent_schema: oa_types.Schema,
    schemas: oa_types.Schemas
) -> str:
    """Calculate the foreign key property."""
    column_name = oa_helpers.foreign_key.calculate_column_name(
        type_=relationship_type,
        property_schema=schema,
        schemas=schemas,
    )
    target_schema = oa_helpers.foreign_key.get_target_schema(
        type_=relationship_type,
        parent_schema=parent_schema,
        property_schema=schema,
        schemas=schemas,
    )
    return oa_helpers.foreign_key.calculate_prop_name(
        type_=relationship_type,
        column_name=column_name,
        property_name=property_name,
        target_schema=target_schema,
        schemas=schemas,
    )


def _get_nullable(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> typing.Optional[bool]:
    """Retrieve the nullable value from a schema."""
    return oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.nullable, schema=schema, schemas=schemas
    )


def _get_many_to_one(
    *,
    property_name: str,
    schema: oa_types.Schema,
    parent_schema: oa_types.Schema,
    schemas: oa_types.Schemas
):
    """
    Retrieve the artifacts for a many-to-one relationship property.

    Args:
        schemas: All the defined schemas.
        property_name: The name of the property.
        schema: The schema of the relationship property to gather artifacts for.
        parent_schema: The schema that contains the property.

    Returns:
        The artifacts for the property.

    """
    sub_type: oa_types.Literal[
        oa_types.RelationshipType.MANY_TO_ONE
    ] = oa_types.RelationshipType.MANY_TO_ONE
    parent = _get_parent(schema=schema, schemas=schemas)

    return types.ManyToOneRelationshipPropertyArtifacts(
        type=oa_types.PropertyType.RELATIONSHIP,
        sub_type=sub_type,
        schema=_calculate_x_to_one_schema(
            parent=parent, schema=schema, schemas=schemas
        ),
        required=False,  # to be fixed on calling function
        parent=parent,
        backref_property=_get_backref_property(schema=schema, schemas=schemas),
        kwargs=_get_kwargs(parent=parent, schema=schema, schemas=schemas),
        write_only=_get_write_only(parent=parent, schema=schema, schemas=schemas),
        description=_get_description(parent=parent, schema=schema, schemas=schemas),
        foreign_key=_get_foreign_key(
            relationship_type=sub_type,
            schema=schema,
            parent_schema=parent_schema,
            schemas=schemas,
        ),
        foreign_key_property=_get_foreign_key_property(
            relationship_type=sub_type,
            property_name=property_name,
            schema=schema,
            parent_schema=parent_schema,
            schemas=schemas,
        ),
        nullable=_get_nullable(schema=schema, schemas=schemas),
    )


def _get_one_to_one(
    *,
    property_name: str,
    schema: oa_types.Schema,
    parent_schema: oa_types.Schema,
    schemas: oa_types.Schemas
):
    """
    Retrieve the artifacts for a one-to-one relationship property.

    Args:
        schemas: All the defined schemas.
        property_name: The name of the property.
        schema: The schema of the relationship property to gather artifacts for.
        parent_schema: The schema that contains the property.

    Returns:
        The artifacts for the property.

    """
    sub_type: oa_types.Literal[
        oa_types.RelationshipType.ONE_TO_ONE
    ] = oa_types.RelationshipType.ONE_TO_ONE
    parent = _get_parent(schema=schema, schemas=schemas)

    return types.OneToOneRelationshipPropertyArtifacts(
        type=oa_types.PropertyType.RELATIONSHIP,
        sub_type=sub_type,
        schema=_calculate_x_to_one_schema(
            parent=parent, schema=schema, schemas=schemas
        ),
        required=False,  # to be fixed on calling function
        parent=parent,
        backref_property=_get_backref_property(schema=schema, schemas=schemas),
        kwargs=_get_kwargs(parent=parent, schema=schema, schemas=schemas),
        write_only=_get_write_only(parent=parent, schema=schema, schemas=schemas),
        description=_get_description(parent=parent, schema=schema, schemas=schemas),
        foreign_key=_get_foreign_key(
            relationship_type=sub_type,
            schema=schema,
            parent_schema=parent_schema,
            schemas=schemas,
        ),
        foreign_key_property=_get_foreign_key_property(
            relationship_type=sub_type,
            property_name=property_name,
            schema=schema,
            parent_schema=parent_schema,
            schemas=schemas,
        ),
        nullable=_get_nullable(schema=schema, schemas=schemas),
    )


def _calculate_one_to_x_schema(
    *, parent: str, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> oa_types.ArrayRefSchema:
    """Calculate the schema for a x-to-one relationship."""
    return_schema: oa_types.ArrayRefSchema = {
        "type": "array",
        "items": {"type": "object", "x-de-$ref": parent},
    }

    description = _get_description(schema=schema, schemas=schemas, parent=parent)
    if description is not None:
        return_schema["description"] = description
    write_only = oa_helpers.peek.write_only(schema=schema, schemas=schemas)
    if write_only is not None:
        return_schema["writeOnly"] = write_only

    return return_schema


def _get_one_to_many(
    *,
    property_name: str,
    schema: oa_types.Schema,
    parent_schema: oa_types.Schema,
    schemas: oa_types.Schemas
):
    """
    Retrieve the artifacts for a one-to-many relationship property.

    Args:
        schemas: All the defined schemas.
        property_name: The name of the property.
        schema: The schema of the relationship property to gather artifacts for.
        parent_schema: The schema that contains the property.

    Returns:
        The artifacts for the property.

    """
    sub_type: oa_types.Literal[
        oa_types.RelationshipType.ONE_TO_MANY
    ] = oa_types.RelationshipType.ONE_TO_MANY
    items_schema = oa_helpers.peek.items(schema=schema, schemas=schemas)
    assert items_schema is not None

    parent = _get_parent(schema=items_schema, schemas=schemas)

    return types.OneToManyRelationshipPropertyArtifacts(
        type=oa_types.PropertyType.RELATIONSHIP,
        sub_type=sub_type,
        schema=_calculate_one_to_x_schema(
            parent=parent, schema=schema, schemas=schemas
        ),
        required=False,  # to be fixed on calling function
        parent=parent,
        backref_property=_get_backref_property(schema=items_schema, schemas=schemas),
        kwargs=_get_kwargs(parent=parent, schema=items_schema, schemas=schemas),
        write_only=_get_write_only(parent=parent, schema=schema, schemas=schemas),
        description=_get_description(parent=parent, schema=schema, schemas=schemas),
        foreign_key=_get_foreign_key(
            relationship_type=sub_type,
            schema=schema,
            parent_schema=parent_schema,
            schemas=schemas,
        ),
        foreign_key_property=_get_foreign_key_property(
            relationship_type=sub_type,
            property_name=property_name,
            schema=schema,
            parent_schema=parent_schema,
            schemas=schemas,
        ),
    )


def _get_secondary(*, schema: oa_types.Schema, schemas: oa_types.Schemas) -> str:
    """Retrieve the secondary from an object reference."""
    secondary = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.secondary, schema=schema, schemas=schemas
    )
    assert secondary is not None

    return secondary


def _get_many_to_many(*, schema: oa_types.Schema, schemas: oa_types.Schemas, **_):
    """
    Retrieve the artifacts for a many-to-many relationship property.

    Args:
        schemas: All the defined schemas.
        property_name: The name of the property.
        schema: The schema of the relationship property to gather artifacts for.
        parent_schema: The schema that contains the property.

    Returns:
        The artifacts for the property.

    """
    items_schema = oa_helpers.peek.items(schema=schema, schemas=schemas)
    assert items_schema is not None

    parent = _get_parent(schema=items_schema, schemas=schemas)

    return types.ManyToManyRelationshipPropertyArtifacts(
        type=oa_types.PropertyType.RELATIONSHIP,
        sub_type=oa_types.RelationshipType.MANY_TO_MANY,
        schema=_calculate_one_to_x_schema(
            parent=parent, schema=schema, schemas=schemas
        ),
        required=False,  # to be fixed on calling function
        parent=parent,
        backref_property=_get_backref_property(schema=items_schema, schemas=schemas),
        kwargs=_get_kwargs(parent=parent, schema=items_schema, schemas=schemas),
        write_only=_get_write_only(parent=parent, schema=schema, schemas=schemas),
        description=_get_description(parent=parent, schema=schema, schemas=schemas),
        secondary=_get_secondary(schema=items_schema, schemas=schemas),
    )


_GET_MAPPING: typing.Dict[oa_types.RelationshipType, typing.Callable] = {
    oa_types.RelationshipType.MANY_TO_ONE: _get_many_to_one,
    oa_types.RelationshipType.ONE_TO_ONE: _get_one_to_one,
    oa_types.RelationshipType.ONE_TO_MANY: _get_one_to_many,
    oa_types.RelationshipType.MANY_TO_MANY: _get_many_to_many,
}


def get(
    schemas: oa_types.Schemas,
    parent_schema: oa_types.Schema,
    property_name: str,
    schema: oa_types.Schema,
    required: bool,
) -> types.TAnyRelationshipPropertyArtifacts:
    """
    Retrieve the artifacts for a relationship property.

    Args:
        schemas: All the defined schemas.
        parent_schema: The schema that contains the property.
        property_name: The name of the property.
        schema: The schema of the relationship property to gather artifacts for.
        required: WHether the property appears in the required list.

    Returns:
        The artifacts for the property.

    """
    sub_type = oa_helpers.relationship.calculate_type(schema=schema, schemas=schemas)

    artifacts = _GET_MAPPING[sub_type](
        property_name=property_name,
        schema=schema,
        schemas=schemas,
        parent_schema=parent_schema,
    )
    artifacts.required = required

    return artifacts



"""Helpers for association tables."""

import typing

from ... import helpers as oa_helpers
from ... import types
from .. import helpers


def _requires_association(schemas: types.Schemas, schema: types.Schema) -> bool:
    """
    Calculate whether a property requires an association table.

    Args:
        schema: The schema of the property to check.

    Returns:
        Whether the property requires an association table.

    """
    return oa_helpers.relationship.is_relationship_type(
        type_=types.RelationshipType.MANY_TO_MANY, schema=schema, schemas=schemas
    )


class TParentPropertySchema(typing.NamedTuple):
    """Holds information about an association table."""

    parent: types.TNameSchema
    property: types.TNameSchema


def get_association_property_iterator(
    *, schemas: types.Schemas
) -> typing.Iterable[TParentPropertySchema]:
    """
    Get an iterator for properties that require association tables from the schemas.

    Assume that the schemas are individually valid.

    To ensure no duplication, property iteration stays within the model context.

    Args:
        schemas: All defined schemas.

    Returns:
        An iterator with properties that require an association table along with the
        schemas and the parent schema.

    """
    constructables = helpers.iterate.constructable(schemas=schemas)
    for name, schema in constructables:
        properties = helpers.iterate.properties_items(
            schema=schema, schemas=schemas, stay_within_model=True
        )
        association_property_schemas = filter(
            lambda args: _requires_association(schemas, args[1]), properties
        )
        yield from (
            TParentPropertySchema(
                parent=types.TNameSchema(name=name, schema=schema),
                property=types.TNameSchema(name=property_name, schema=property_schema),
            )
            for property_name, property_schema in association_property_schemas
        )


def get_secondary(*, schema: types.Schema, schemas: types.Schemas) -> str:
    """
    Get the secondary value from a many-to-many relationship property.

    Assume that the property is a valid many-to-many relationship.

    Args:
        schema: The schema of the many-to-many property.
        schemas: All defined schemas.

    Returns:
        The value of x-secondary.

    """
    items = oa_helpers.peek.items(schema=schema, schemas=schemas)
    assert items is not None
    secondary = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.secondary, schema=items, schemas=schemas
    )
    assert secondary is not None and isinstance(secondary, str)
    return secondary


def get_secondary_parent_property_schema_mapping(
    *, schemas: types.Schemas
) -> typing.Dict[str, TParentPropertySchema]:
    """
    Get a mapping keyed with the x-secondary value and the parent and property schema.

    Assume that the schemas are individually valid.

    To ensure no duplication, property iteration stays within the model context.

    Args:
        schemas: All defined schemas.

    Returns:
        A dictionary with the x-secondary value is the key and the parent and property
        schema of an association as the value.

    """
    association_properties = get_association_property_iterator(schemas=schemas)
    association_name_parent_property_schemas = map(
        lambda association: (
            get_secondary(schema=association.property.schema, schemas=schemas),
            association,
        ),
        association_properties,
    )
    return dict(association_name_parent_property_schemas)


class TCalculatePropertySchemaReturn(typing.NamedTuple):
    """The return type for the calculate_property_schema function."""

    name: str
    schema: types.ColumnSchema


def calculate_property_schema(
    *, schema: types.Schema, schemas: types.Schemas
) -> TCalculatePropertySchemaReturn:
    """
    Calculate the property name and schema for a column for a many-to-many relationship.

    Assume that all schemas are valid.
    Assume that the schema has a single primary key.

    The following algorithm is used to calculate key artifacts:
    - the type is the type of the primary key property of the schema
    - if defined on the primary key property, the format is the format of the primary
        key property
    - if defined on the primary key property, the maxLength is the maxLength of the
        primary key property
    - the primary key is set to True
    - the foreign key is <schema tablename>.<primary key property name>
    - the property name is <schema tablename>_<primary key property name>

    Args:
        schemas: All defined schemas.
        schema: The model schema to use to calculate the property schema.

    Returns:
        The property name and schema for one of the columns of the association table
        that handles one of the sides of the many to many relationship.

    """
    # Retrieve primary key column
    properties = helpers.iterate.properties_items(
        schema=schema, schemas=schemas, stay_within_tablename=True
    )
    primary_key_properties = filter(
        lambda args: oa_helpers.peek.primary_key(schema=args[1], schemas=schemas)
        is True,
        properties,
    )
    primary_key_property = next(primary_key_properties, None)
    assert primary_key_property is not None
    assert next(primary_key_properties, None) is None

    # Get artifacts
    tablename = oa_helpers.peek.prefer_local(
        get_value=oa_helpers.peek.tablename, schema=schema, schemas=schemas
    )
    primary_key_property_name, primary_key_property_schema = primary_key_property
    type_ = oa_helpers.peek.type_(schema=primary_key_property_schema, schemas=schemas)
    format_ = oa_helpers.peek.format_(
        schema=primary_key_property_schema, schemas=schemas
    )
    max_length = oa_helpers.peek.max_length(
        schema=primary_key_property_schema, schemas=schemas
    )
    foreign_key = oa_helpers.foreign_key.calculate_foreign_key(
        column_name=primary_key_property_name, target_schema=schema, schemas=schemas
    )

    # Calculate name and schema
    property_name = f"{tablename}_{primary_key_property_name}"
    property_schema: types.ColumnSchema = {
        "type": type_,
        "x-primary-key": True,
        "x-foreign-key": foreign_key,
    }
    if format_ is not None:
        property_schema["format"] = format_
    if max_length is not None:
        property_schema["maxLength"] = max_length

    return TCalculatePropertySchemaReturn(name=property_name, schema=property_schema)


def calculate_schema(
    *,
    property_schema: types.Schema,
    parent_schema: types.Schema,
    schemas: types.Schemas,
) -> types.TNameSchema:
    """
    Calculate the schema for the association table.

    Assume that all schemas are valid.
    Assume that the property schema is a many-to-many relationship.

    The following algorithm is used to calculate the schema:
    1. retrieve the secondary value,
    2. calculate the property schema for the parent schema,
    3. retrieve the referenced schema and calculate the property schema,
    4. assemble the final schema with
        a. the object type,
        b. tablename of the secondary value,
        c. properties based on the parent and referenced schema and
        d. required with the property names and
    5. calculate the schema name based on the secondary value and avoiding any existing
        schema names.

    The following is the form of the schema:
    <association schema name>:
        type: object
        x-tablename: <x-secondary>
        properties:
            <parent tablename>_<parent property name>:
                type: <parent property type>
                x-primary-key: true
                # if the parent property defines a format
                format: <parent property format>
                # if the parent property defines maxLength
                maxLength: <parent property maxLength>
                x-foreign-key: <parent tablename>.<parent property name>
            <child tablename>_<child property name>:
                type: <child property type>
                x-primary-key: true
                # if the child property defines a format
                format: <child property format>
                # if the child property defines maxLength
                maxLength: <child property maxLength>
                x-foreign-key: <child tablename>.<child property name>
        required:
            - <parent tablename>_<parent property name>
            - <child tablename>_<child property name>

    Args:
        parent_schema: The schema the property is embedded in.
        property_schema: The schema of the many-to-many property.
        schemas: All defined schemas.

    Returns:
        The schema of the association table.

    """
    secondary = get_secondary(schema=property_schema, schemas=schemas)
    parent_property = calculate_property_schema(schema=parent_schema, schemas=schemas)
    _, ref_schema = oa_helpers.relationship.get_ref_schema_many_to_x(
        property_schema=property_schema, schemas=schemas
    )
    ref_property = calculate_property_schema(schema=ref_schema, schemas=schemas)

    schema = {
        "type": "object",
        "x-tablename": secondary,
        "properties": {
            parent_property.name: parent_property.schema,
            ref_property.name: ref_property.schema,
        },
        "required": [parent_property.name, ref_property.name],
    }

    name = secondary.title().replace("_", "")
    while name in schemas:
        name = f"Autogen{name}"

    return types.TNameSchema(name=name, schema=schema)
